import tensorflow as tf

from TFGENZOO.flows.flowbase import FlowComponent


class Actnorm(FlowComponent):
    """Actnorm Layer

    This layer can SyncBatch Normalization, but may crash frequently.

    Sources:

        https://github.com/openai/glow/blob/master/tfops.py#L71-L163

    Note:

        * initialize
            | mean = mean(first_batch)
            | var = variance(first_batch)
            | logs = log(scale / sqrt(var)) / logscale_factor
            | bias = - mean

        * forward formula
            | logs = logs * logscale_factor
            | scale = exp(logs)
            z = (x + bias) * scale
            log_det_jacobain = sum(logs) * H * W


        * inverse formula
            | logs = logs * logsscale_factor
            | inv_scale = exp(-logs)
            | z = x * inv_scale - bias
            | inverse_log_det_jacobian = sum(- logs) * H * W

    Attributes:
        calc_ldj: bool
            flag of calculate log det jacobian
        scale: float
            initialize batch's variance scaling
        logscale_factor: float
            barrier log value to - Inf
    """

    def __init__(self, scale: float = 1.0, logscale_factor: float = 3.0, **kwargs):
        super(Actnorm, self).__init__(**kwargs)
        self.scale = scale
        self.logscale_factor = logscale_factor

    def get_config(self):
        config = super().get_config()
        config_update = {"scale": self.scale, "logscale_factor": self.logscale_factor}
        config.update(config_update)
        return config

    # pylint: disable=attribute-defined-outside-init
    def build(self, input_shape: tf.TensorShape):
        if len(input_shape) == 4:
            reduce_axis = [0, 1, 2]
            b, h, w, c = list(input_shape)
            self.logdet_factor = tf.constant(h * w, dtype=tf.float32)
        else:
            raise NotImplementedError()

        self.reduce_axis = reduce_axis

        # stats_shape = [1, 1, 1, C] if input_shape == [B, H, W, C]
        stats_shape = [1 for _ in range(len(input_shape))]
        stats_shape[-1] = input_shape[-1]

        # self.logs = self.add_weight(
        #     name="logscale",
        #     shape=tuple(stats_shape),
        #     initializer="zeros",
        #     trainable=True,
        # )
        # self.bias = self.add_weight(
        #     name="bias", shape=tuple(stats_shape), initializer="zeros", trainable=True
        # )

        self.mean = self.add_weight(
            name="mean",
            shape=tuple(stats_shape),
            initializer="zeros",
            trainable=True,
            aggregation=tf.VariableAggregation.MEAN,
        )
        self.squared = self.add_weight(
            name="squared",
            shape=tuple(stats_shape),
            initializer="zeros",
            trainable=True,
            aggregation=tf.VariableAggregation.MEAN,
        )

        super().build(input_shape)

    @property
    def bias(self):
        return -self.mean

    @property
    def logs(self):
        # var(x) = E(x^2) - E(x)^2
        variance = self.squared - tf.square(self.mean)
        logs = (
            #     # var(x) = E(x^2) - E(x)^2
            tf.math.log(self.scale * tf.math.rsqrt(variance + 1e-6))
            / self.logscale_factor
            #     variance = self.squared - tf.square(self.mean)
        )
        return logs

    def data_dep_initialize(self, x: tf.Tensor):

        if self.initialized:
            # bias, logs = self.bias, self.logs
            mean = self.mean
            squared = self.squared
        else:
            tf.print("initialization at {}".format(self.name))
            # mean, variance = tf.nn.moments(x, axes=[0, 1, 2], keepdims=True)
            # bias = -mean
            # logs = (
            #     #     # var(x) = E(x^2) - E(x)^2
            #     tf.math.log(self.scale * tf.math.rsqrt(variance + 1e-6))
            #     / self.logscale_factor
            #     #     variance = self.squared - tf.square(self.mean)
            # )
            mean = tf.reduce_mean(x, axis=[0, 1, 2], keepdims=True)
            squared = tf.reduce_mean(tf.square(x), axis=[0, 1, 2], keepdims=True)

        with tf.control_dependencies([mean, squared]):
            self.mean.assign(mean)
            self.squared.assign(squared)
            # self.bias.assign(bias)
            # self.logs.assign(logs)

            super().data_dep_initialize(x)

    def forward(self, x: tf.Tensor, **kwargs):
        logs = self.logs * self.logscale_factor

        # centering
        z = x + self.bias
        # scaling
        z = z * tf.exp(logs)

        log_det_jacobian = tf.reduce_sum(logs) * self.logdet_factor
        log_det_jacobian = tf.broadcast_to(log_det_jacobian, tf.shape(x)[0:1])
        return z, log_det_jacobian

    def inverse(self, z: tf.Tensor, **kwargs):
        logs = self.logs * self.logscale_factor

        # inverse scaling
        x = z * tf.exp(-1 * logs)
        # inverse centering
        x = x - self.bias

        inverse_log_det_jacobian = -1 * tf.reduce_sum(logs) * self.logdet_factor
        inverse_log_det_jacobian = tf.broadcast_to(
            inverse_log_det_jacobian, tf.shape(x)[0:1]
        )
        return x, inverse_log_det_jacobian



from TFGENZOO.flows.utils.conv import Conv2D
from TFGENZOO.flows.utils.conv_zeros import Conv2DZeros, Conv1DZeros
from TFGENZOO.flows.utils.actnorm_activation import ActnormActivation
from TFGENZOO.flows.utils.gaussianize import gaussian_likelihood, gaussian_sample
from TFGENZOO.flows.utils.util import bits_x, split_feature


__all__ = [
    "Conv2D",
    "Conv2DZeros",
    "Conv1DZeros",
    "ActnormActivation",
    "gaussian_likelihood",
    "gaussian_sample",
    "bits_x",
    "split_feature",
]



from typing import Tuple

import tensorflow as tf
from tensorflow.keras import layers

Layer = layers.Layer


class Conv2DZeros(Layer):
    """Convolution layer for NHWC image with zero initialization
    Sources:
   
        https://github.com/openai/glow/blob/master/tfops.py#L292-L313

    Note:
        this layer not implemented.
        * function add_edge_padding
           ref. https://github.com/openai/glow/blob/master/tfops.py#L203-L232

        Xavier Initialize is better than other initializer

            ref. https://arxiv.org/pdf/1907.02392.pdf

            code.

                * https://www.tensorflow.org/api_docs/python/tf/keras/initializers/GlorotNormal
                * https://www.tensorflow.org/api_docs/python/tf/keras/initializers/GlorotUniform?version=nightly
    """

    def __init__(
        self,
        width: int = None,
        width_scale: int = 1,
        kernel_size: Tuple[int, int] = (3, 3),
        stride: Tuple[int, int] = (1, 1),
        padding: str = "SAME",
        logscale_factor: float = 3.0,
        initializer: tf.keras.initializers.Initializer = "zeros",
    ):
        super().__init__()
        self.width = width
        self.width_scale = width_scale
        self.kernel_size = list(kernel_size)
        self.stride = [1] + list(stride) + [1]
        self.padding = padding
        self.logscale_factor = logscale_factor
        self.initializer = initializer

    def get_config(self):
        config = super().get_config()
        config_update = {
            "width": self.width,
            "width_scale": self.width_scale,
            "kernel_size": self.kernel_size,
            "stride": self.stride,
            "padding": self.padding,
            "log_scale_factor": self.logscale_factor,
            "initializer": tf.keras.initializers.serialize(
                tf.keras.initializers.get(self.initializer)
            ),
        }
        config.update(config_update)
        return config

    def build(self, input_shape: tf.TensorShape):
        n_in = input_shape[-1]
        n_out = self.width if self.width is not None else n_in * self.width_scale
        filters = self.kernel_size + [n_in, n_out]
        self.kernel = self.add_weight(
            name="kernel", initializer=self.initializer, shape=filters, dtype=tf.float32
        )
        self.bias = self.add_weight(
            name="bias",
            shape=[1 for _ in range(len(input_shape) - 1)] + [n_out],
            initializer="zeros",
        )
        self.logs = self.add_weight(
            name="logs", shape=[1, n_out], initializer=self.initializer
        )
        super().build(input_shape)

    def call(self, x: tf.Tensor):
        x = tf.nn.conv2d(
            x, filters=self.kernel, strides=self.stride, padding=self.padding
        )
        x += self.bias
        x *= tf.exp(self.logs * self.logscale_factor)
        return x


class Conv1DZeros(tf.keras.layers.Layer):
    """Convolution layer for NTC text/audio with zero initialization

    Sources:

        https://github.com/openai/glow/blob/master/tfops.py#L292-L313

    Note:
        this layer not implemented.
        * function add_edge_padding
           ref. https://github.com/openai/glow/blob/master/tfops.py#L203-L232

        Xavier Initialize is better than other initializer

            ref. https://arxiv.org/pdf/1907.02392.pdf

            code.

                * https://www.tensorflow.org/api_docs/python/tf/keras/initializers/GlorotNormal
                * https://www.tensorflow.org/api_docs/python/tf/keras/initializers/GlorotUniform?version=nightly

    Examples:

        >>> import tensorflow as tf
        >>> from TFGENZOO.flows.utils.conv_zeros import Conv1DZeros
        >>> c1z = Conv1DZeros(width_scale = 2)
        >>> x = tf.keras.layers.Input([None, 32]) # [B, T, C] where T is time-step and C is hidden-depth
        >>> y = clz(y) # [B, T, C * 2]
    """

    def __init__(
        self,
        width: int = None,
        width_scale: int = 2,
        kernel_size: int = 3,
        stride: int = 1,
        padding: str = "SAME",
        logscale_factor: float = 3.0,
        initializer: tf.keras.initializers.Initializer = "zeros",
    ):
        super().__init__()
        self.width = width
        self.width_scale = width_scale
        self.kernel_size = kernel_size
        self.stride = stride
        self.padding = padding
        self.initializer = initializer
        self.logscale_factor = logscale_factor

    def get_config(self):
        config = super().get_config()
        config_update = {
            "width_scale": self.width_scale,
            "initializer": tf.keras.initializers.serialize(
                tf.keras.initializers.get(self.initializer)
            ),
            "kernel_size": self.kernel_size,
            "stride": self.stride,
            "padding": self.padding,
        }
        config.update(config_update)
        return config

    def build(self, input_shape: tf.TensorShape):
        n_in = input_shape[-1]
        n_out = self.width if self.width is not None else n_in * self.width_scale
        self.kernel = self.add_weight(
            name="kernel",
            initializer=self.initializer,
            shape=[self.kernel_size] + [n_in, n_out],
            dtype=tf.float32,
        )
        self.bias = self.add_weight(
            name="bias",
            shape=[1 for _ in range(len(input_shape) - 1)] + [n_out],
            initializer="zeros",
        )
        self.logs = self.add_weight(
            name="logs", shape=[1, n_out], initializer=self.initializer
        )
        super().build(input_shape)

    def call(self, x: tf.Tensor):
        x = tf.nn.conv1d(
            x, filters=self.kernel, stride=self.stride, padding=self.padding
        )
        x += self.bias
        x *= tf.exp(self.logs * self.logscale_factor)
        return x



import numpy as np
import tensorflow as tf
import tensorflow_probability as tfp


def gaussian_likelihood(mean: tf.Tensor, logsd: tf.Tensor, x: tf.Tensor):
    r"""calculate negative log likelihood of Gaussian Distribution.

    Args:
        mean  (tf.Tensor): mean [B, ...]
        logsd (tf.Tensor): log standard deviation [B, ...]
        x     (tf.Tensor): tensor [B, ...]

    Returns:
        ll    (tf.Tensor): log likelihood [B, ...]

    Note:

       .. math::
          :nowrap:

          \begin{align}
          ll &= - \cfrac{1}{2} (k\log(2 \pi) + \log |Var| \\
             &+ (x - Mu)^T (Var ^ {-1}) (x - Mu))\\
          ,\ where & \\
               & k = 1\ (Independent)\\
               & Var\ is\ a\ variance = exp(2 logsd)
          \end{align}
    """
    c = np.log(2 * np.pi)
    ll = -0.5 * (c + 2.0 * logsd + ((x - mean) ** 2) / tf.math.exp(2.0 * logsd))
    return ll


def gaussian_sample(mean: tf.Tensor, logsd: tf.Tensor, temparature: float = 1.0):
    r"""sampling from mean, logsd * temparature

    Args:
        mean (tf.Tensor): mean [B, ...]
        logsd(tf.Tensor): log standard deviation [B, ...]
        temparature      (float): temparature

    Returns:
        new_z(tf.Tensor): sampled latent variable [B, ...]

    Noto:
        I cann't gurantee it's correctness.
        Please open the tensorflow probability's Issue.
    """
    return tfp.distributions.Normal(mean, tf.math.exp(logsd) * temparature).sample()



from enum import Enum
from typing import Callable

import tensorflow as tf
from tensorflow.keras import layers

from TFGENZOO.flows.flowbase import FlowComponent

Layer = layers.Layer
Conv2D = layers.Conv2D


class LogScale(Layer):
    def build(self, input_shape: tf.TensorShape):
        shape = [1, input_shape[-1]]
        self.logs = self.add_weight(
            name="log_scale", shape=tuple(shape), initializer="zeros", trainable=True
        )

    def __init__(self, log_scale_factor: float = 3.0, **kwargs):
        super(LogScale, self).__init__(**kwargs)
        self.log_scale_factor = log_scale_factor

    def get_config(self):
        config = super().get_config()
        config_update = {"log_scale_factor": self.log_scale_factor}
        config.update(config_update)
        return config

    def call(self, x: tf.Tensor):
        return x * tf.exp(self.logs * self.log_scale_factor)


class AffineCouplingMask(Enum):
    ChannelWise = 1


class AffineCoupling(FlowComponent):
    """Affine Coupling Layer

    Sources:
        https://github.com/masa-su/pixyz/blob/master/pixyz/flows/coupling.py

    Note:
        * forward formula
            | [x1, x2] = split(x)
            | log_scale, shift = NN(x1)
            | scale = sigmoid(log_scale + 2.0)
            | z1 = x1
            | z2 = (x2 + shift) * scale
            | z = concat([z1, z2])
            | LogDetJacobian = sum(log(scale))

        * inverse formula
            | [z1, z2] = split(x)
            | log_scale, shift = NN(z1)
            | scale = sigmoid(log_scale + 2.0)
            | x1 = z1
            | x2 = z2 / scale - shift
            | z = concat([x1, x2])
            | InverseLogDetJacobian = - sum(log(scale))

        * implementation notes
           | in Glow's Paper, scale is calculated by exp(log_scale),
           | but IN IMPLEMENTATION, scale is done by sigmoid(log_scale + 2.0)

    Examples:

        >>> import tensorflow as tf
        >>> from TFGENZOO.flows.affine_coupling import AffineCoupling
        >>> from TFGENZOO.layers.resnet import ShallowResNet
        >>> af = AffineCoupling(scale_shift_net_template=ShallowResNet)
        >>> af.build([None, 16, 16, 4])
        >>> af.get_config()
            {'name': 'affine_coupling_1', ...}
        >>> inputs = tf.keras.Input([16, 16, 4])
        >>> af(inputs)
        (<tf.Tensor 'affine_coupling_3_2/Identity:0' shape=(None, 16, 16, 4) dtype=float32>,
         <tf.Tensor 'affine_coupling_3_2/Identity_1:0' shape=(None,) dtype=float32>)
        >>> tf.keras.Model(inputs, af(inputs)).summary()
        Model: "model_1"
        _________________________________________________________________
        Layer (type)                 Output Shape              Param #
        =================================================================
        input_3 (InputLayer)         [(None, 16, 16, 4)]       0
        _________________________________________________________________
        affine_coupling (AffineCoupl ((None, 16, 16, 4), (None 2389003
        =================================================================
        Total params: 2,389,003
        Trainable params: 0
        Non-trainable params: 2,389,003
        _________________________________________________________________
    """

    def __init__(
        self,
        mask_type: AffineCouplingMask = AffineCouplingMask.ChannelWise,
        scale_shift_net: Layer = None,
        scale_shift_net_template: Callable[[tf.keras.Input], tf.keras.Model] = None,
        scale_type="safe_exp",
        **kwargs
    ):
        """
        Args:
            mask_type       (AffineCouplingMask: AffineCoupling Mask type
            scale_shift_net (tf.keras.Layer): NN in the fomula (Deprecated)
            scale_shift_net_template (Callable[[tf.keras.Input], [tf.keras.Model]]): NN in the formula (for tf.keras.Model without Input Shape)
        """
        super().__init__(**kwargs)
        if not scale_shift_net_template and not scale_shift_net:
            raise ValueError
        if scale_shift_net_template is not None:
            self.scale_shift_net_template = scale_shift_net_template
            self.scale_shift_net = None
        elif scale_shift_net is not None:
            self.scale_shift_net = scale_shift_net
        self.mask_type = mask_type

        self.scale_type = scale_type
        if self.scale_type not in ["safe_exp", "exp", "sigmoid"]:
            raise ValueError
        if self.scale_type == "safe_exp":
            self.scale_func = lambda log_scale: tf.exp(
                tf.clip_by_value(log_scale, -15.0, 15.0)
            )
        elif self.scale_type == "exp":
            self.scale_func = lambda log_scale: tf.exp(log_scale)
        else:
            self.scale_func = lambda log_scale: tf.nn.sigmoid(log_scale + 2.0)

    def get_config(self):
        config = super().get_config()
        config_update = {
            "scale_shit_net": self.scale_shift_net.get_config(),
            "mask_type": self.mask_type,
            "scale_type": self.scale_type,
        }
        config.update(config_update)
        return config

    def build(self, input_shape: tf.TensorShape):
        self.reduce_axis = list(range(len(input_shape)))[1:]
        if self.scale_shift_net is None:
            resnet_inputs = list(input_shape)[1:]
            resnet_inputs[-1] = int(resnet_inputs[-1] / 2)
            self.scale_shift_net = self.scale_shift_net_template(
                tf.keras.Input(resnet_inputs)
            )

        super().build(input_shape)

    def forward(self, x: tf.Tensor, **kwargs):
        x1, x2 = tf.split(x, 2, axis=-1)
        z1 = x1
        h = self.scale_shift_net(x1, **kwargs)
        if self.mask_type == AffineCouplingMask.ChannelWise:
            shift = h[..., 0::2]
            log_scale = h[..., 1::2]

            scale = self.scale_func(log_scale)
            z2 = (x2 + shift) * scale

            # scale's shape is [B, H, W, C]
            # log_det_jacobian's hape is [B]
            log_det_jacobian = tf.reduce_sum(tf.math.log(scale), axis=self.reduce_axis)
            return tf.concat([z1, z2], axis=-1), log_det_jacobian
        else:
            raise NotImplementedError()

    def inverse(self, z: tf.Tensor, **kwargs):
        z1, z2 = tf.split(z, 2, axis=-1)
        x1 = z1
        h = self.scale_shift_net(z1, **kwargs)
        if self.mask_type == AffineCouplingMask.ChannelWise:
            shift = h[..., 0::2]
            log_scale = h[..., 1::2]

            scale = self.scale_func(log_scale)
            x2 = (z2 / scale) - shift

            # scale's shape is [B, H, W, C // 2]
            # inverse_log_det_jacobian's hape is [B]
            inverse_log_det_jacobian = -1 * tf.reduce_sum(
                tf.math.log(scale), axis=self.reduce_axis
            )
            return tf.concat([x1, x2], axis=-1), inverse_log_det_jacobian
        else:
            raise NotImplementedError()



from typing import Dict

import tensorflow as tf
from tensorflow.keras import Model

from TFGENZOO.flows.actnorm import Actnorm
from TFGENZOO.flows.affine_coupling import AffineCoupling
from TFGENZOO.flows.factor_out import FactorOut
from TFGENZOO.flows.flowbase import FactorOutBase, FlowModule
from TFGENZOO.flows.inv1x1conv import Inv1x1Conv
from TFGENZOO.flows.quantize import LogitifyImage
from TFGENZOO.flows.squeeze import Squeeze
from TFGENZOO.layers.resnet import ShallowResNet


class SingleFlow(Model):
    # super(BasicGlow, self).build(input_shape)
    # 5, 3
    def __init__(self, K: int = 5, L: int = 1, resblk_kwargs: Dict = None):
        super().__init__()
        if resblk_kwargs is None:
            resblk_kwargs = {"num_block": 3, "units_factor": 6}

        self.resblk_kwargs = resblk_kwargs
        self.K = K
        self.L = L
        layers = []
        layers.append(LogitifyImage())
        for _ in range(3):
            layers.append(Squeeze())
            for _ in range(5):
                layers.append(Actnorm())
                layers.append(Inv1x1Conv())
                layers.append(
                    AffineCoupling(scale_shift_net=ShallowResNet(**self.resblk_kwargs))
                )
        self.flows = layers

    def get_config(self):
        config = super().get_config()
        config_update = {
            "resblk_kwargs": self.resblk_kwargs,
            "K": self.K,
            "L": self.L,
            "flows": [layer.get_config() for layer in self.flows],
        }
        config.update(config_update)
        return config

    def call(self, x, zaux=None, inverse=False, training=True):
        if inverse:
            return self.inverse(x, zaux, training)
        else:
            return self.forward(x, training)

    def inverse(self, x, zaux, training):
        inverse_log_det_jacobian = tf.zeros(tf.shape(x)[0:1])
        for flow in reversed(self.flows):
            if isinstance(flow, Squeeze):
                if flow.with_zaux:
                    x, zaux = flow(x, zaux=zaux, inverse=True)
                else:
                    x = flow(x, inverse=True)
            elif isinstance(flow, FactorOutBase):
                if flow.with_zaux:
                    x, zaux = flow(x, zaux=zaux, inverse=True)
                else:
                    x = flow(x, zaux=zaux, inverse=True)
            else:
                x, ldj = flow(x, inverse=True, training=training)
                inverse_log_det_jacobian += ldj
        return x, inverse_log_det_jacobian

    def forward(self, x, training):
        zaux = None
        log_det_jacobian = tf.zeros(tf.shape(x)[0:1])
        log_likelihood = tf.zeros(tf.shape(x)[0:1])
        for flow in self.flows:
            if isinstance(flow, Squeeze):
                if flow.with_zaux:
                    x, zaux = flow(x, zaux=zaux)
                else:
                    x = flow(x)
            elif isinstance(flow, FactorOutBase):
                x, zaux, ll = flow(x, zaux)
                # TODO: testing
                log_likelihood += ll
            else:
                x, ldj = flow(x, training=training)
                log_det_jacobian += ldj
        return x, log_det_jacobian, log_likelihood  # , zaux


class BasicGlow(Model):
    # def build(self, input_shape: tf.TensorShape):
    #     super(BasicGlow, self).build(input_shape)
    # 5, 3
    def __init__(
        self,
        K: int = 5,
        L: int = 3,
        resblk_kwargs: Dict = None,
        conditional: bool = False,
    ):
        super(BasicGlow, self).__init__()
        if resblk_kwargs is None:
            resblk_kwargs = {"num_block": 3, "units_factor": 6}
        self.resblk_kwargs = resblk_kwargs
        self.K = K
        self.L = L
        layers = []
        layers.append(LogitifyImage())
        for l in range(self.L):
            if l == 0:
                layers.append(Squeeze(with_zaux=False))
            else:
                layers.append(Squeeze(with_zaux=True))
            fml = []
            for k in range(self.K):
                fml.append(Actnorm())
                fml.append(Inv1x1Conv())
                fml.append(
                    AffineCoupling(scale_shift_net=ShallowResNet(**self.resblk_kwargs))
                )
            layers.append(FlowModule(fml))
            if l == 0:
                layers.append(FactorOut(conditional=conditional))
            elif l != self.L - 1:
                layers.append(FactorOut(with_zaux=True, conditional=conditional))
        self.flows = layers

    def get_config(self):
        config = super().get_config()
        config_update = {
            "resblk_kwargs": self.resblk_kwargs,
            "K": self.K,
            "L": self.L,
            "flows": [layer.get_config() for layer in self.flows],
        }
        config.update(config_update)
        return config

    def call(self, x, zaux=None, inverse=False, training=True):
        if inverse:
            return self.inverse(x, zaux, training)
        else:
            return self.forward(x, training)

    def inverse(self, x, zaux, training):
        inverse_log_det_jacobian = tf.zeros(tf.shape(x)[0:1])
        for flow in reversed(self.flows):
            if isinstance(flow, Squeeze):
                if flow.with_zaux:
                    x, zaux = flow(x, zaux=zaux, inverse=True)
                else:
                    x = flow(x, inverse=True)
            elif isinstance(flow, FactorOutBase):
                if flow.with_zaux:
                    x, zaux = flow(x, zaux=zaux, inverse=True)
                else:
                    x = flow(x, zaux=zaux, inverse=True)
            else:
                x, ldj = flow(x, inverse=True, training=training)
                inverse_log_det_jacobian += ldj
        return x, inverse_log_det_jacobian

    def forward(self, x, training):
        zaux = None
        log_det_jacobian = tf.zeros(tf.shape(x)[0:1])
        log_likelihood = tf.zeros(tf.shape(x)[0:1])
        for flow in self.flows:
            if isinstance(flow, Squeeze):
                if flow.with_zaux:
                    x, zaux = flow(x, zaux=zaux)
                else:
                    x = flow(x)
            elif isinstance(flow, FactorOutBase):
                x, zaux, ll = flow(x, zaux=zaux)
                log_likelihood += ll
            else:
                x, ldj = flow(x, training=training)
                log_det_jacobian += ldj
        return x, log_det_jacobian, zaux, log_likelihood


class SqueezeFactorOut(Model):
    def __init__(self, L=3):
        super(SqueezeFactorOut, self).__init__()
        self.flows = [
            LogitifyImage(),
            Squeeze(),
            FactorOut(),
            Squeeze(with_zaux=True),
            FactorOut(with_zaux=True),
            Squeeze(with_zaux=True),
            FactorOut(with_zaux=True),
        ]

    def get_config(self):
        config = super().get_config()
        config_update = {"flows": [flow.get_config() for flow in self.flows]}
        config.update(config_update)
        return config

    def call(self, x, zaux=None, inverse=False):
        if inverse:
            return self.inverse(x, zaux)
        else:
            return self.forward(x)

    def inverse(self, x, zaux):
        inverse_log_det_jacobian = tf.zeros(tf.shape(x)[0:1])
        for flow in reversed(self.flows):
            if isinstance(flow, Squeeze):
                if flow.with_zaux:
                    x, zaux = flow(x, zaux=zaux, inverse=True)
                else:
                    x = flow(x, inverse=True)
            elif isinstance(flow, FactorOutBase):
                if flow.with_zaux:
                    x, zaux = flow(x, zaux=zaux, inverse=True)
                else:
                    x = flow(x, zaux=zaux, inverse=True)
            else:
                x, ldj = flow(x, inverse=True)
                inverse_log_det_jacobian += ldj
        return x, inverse_log_det_jacobian

    def forward(self, x):
        zaux = None
        log_det_jacobian = tf.zeros(tf.shape(x)[0:1])
        log_likelihood = tf.zeros(tf.shape(x)[0:1])
        for flow in self.flows:
            if isinstance(flow, Squeeze):
                if flow.with_zaux:
                    x, zaux = flow(x, zaux=zaux)
                else:
                    x = flow(x)
            elif isinstance(flow, FactorOutBase):
                x, zaux, _ll = flow(x, zaux)
                log_likelihood += _ll
            else:
                x, ldj = flow(x)
                log_det_jacobian += ldj
        return x, log_det_jacobian, zaux, log_likelihood


def squeeze_factor_Test():
    tf.debugging.enable_check_numerics()
    x = tf.keras.Input([24, 24, 1])
    model = SqueezeFactorOut()
    model.build(x.shape)
    z, ldj, zaux, ll = model(x)
    print(z.shape)
    print(ldj.shape)
    print(ll.shape)
    model.summary()
    x = tf.abs(tf.random.normal([16, 24, 24, 1]))
    x = tf.clip_by_value(x, 0.5, 1.0)
    z, ldj, zaux, ll = model(x, inverse=False)
    print(z.shape)
    print(ldj.shape)
    print(zaux.shape)
    print(ll.shape)

    tf.debugging.disable_check_numerics()
    return model


def basic_glow_Test():
    tf.debugging.enable_check_numerics()
    x = tf.keras.Input([24, 24, 1])
    # model = BasicGlow(conditional=True)
    model = BasicGlow(conditional=False)
    model.build(x.shape)
    z, ldj, zaux, ll = model(x)
    print(z.shape)
    print(ldj.shape)
    print(ll.shape)
    model.summary()

    train, test = tf.keras.datasets.mnist.load_data()
    train_image = train[0] / 255.0
    train_image = train_image[..., tf.newaxis]
    train_image = tf.compat.v1.image.resize_bilinear(train_image, size=(24, 24))
    # forward -> inverse
    train_image = train_image[0:12]
    forward, ldj, zaux, ll = model(train_image, inverse=False)
    print("max-min")
    tf.print(tf.reduce_max(forward))
    tf.print(tf.reduce_min(forward))
    tf.print(tf.reduce_sum(tf.cast(tf.math.is_inf(forward), tf.float32)))
    tf.print(tf.reduce_sum(tf.cast(tf.math.is_nan(forward), tf.float32)))
    inverse, ildj = model(forward, zaux=zaux, inverse=True)
    print(forward.shape)
    print(zaux.shape)
    print(ldj.shape)
    print(ildj.shape)
    print("diffs-ldj-rec")
    tf.print(tf.reduce_mean(ldj + ildj))
    tf.print(tf.reduce_mean(train_image - inverse))
    train_image = inverse
    print(train_image.shape)
    import matplotlib.pyplot as plt

    fig = plt.figure(figsize=(18, 18))
    for i in range(9):
        img = tf.squeeze(train_image[i])
        fig.add_subplot(3, 3, i + 1)
        plt.title(train[1][i])
        plt.tick_params(bottom=False, left=False, labelbottom=False, labelleft=False)
        plt.imshow(img, cmap="gray_r")
    plt.show(block=True)

    tf.debugging.disable_check_numerics()
    # return model


def basic_flow_Test():
    tf.debugging.enable_check_numerics()
    x = tf.keras.Input([24, 24, 1])
    model = SingleFlow()
    model.build(x.shape)
    z, ldj = model(x)
    print(z)
    print(ldj)
    print(z.shape)
    print(ldj.shape)
    model.summary()

    train, test = tf.keras.datasets.mnist.load_data()
    train_image = train[0] / 255.0
    train_image = train_image[..., tf.newaxis]
    train_image = tf.compat.v1.image.resize_bilinear(train_image, size=(24, 24))
    # forward -> inverse
    train_image = train_image[0:12]
    forward, ldj = model(train_image, inverse=False)
    print("max-min")
    tf.print(tf.reduce_max(forward))
    tf.print(tf.reduce_min(forward))
    tf.print(tf.reduce_sum(tf.cast(tf.math.is_inf(forward), tf.float32)))
    tf.print(tf.reduce_sum(tf.cast(tf.math.is_nan(forward), tf.float32)))
    inverse, ildj = model(forward, inverse=True)
    print(forward.shape)
    print(ldj.shape)
    print(ildj.shape)
    print("diffs")
    tf.print(tf.reduce_sum(ldj + ildj))
    tf.print(tf.reduce_mean(train_image - inverse))
    train_image = inverse
    print(train_image.shape)
    import matplotlib.pyplot as plt

    fig = plt.figure(figsize=(18, 18))
    for i in range(9):
        img = tf.squeeze(train_image[i])
        fig.add_subplot(3, 3, i + 1)
        plt.title(train[1][i])
        plt.tick_params(bottom=False, left=False, labelbottom=False, labelleft=False)
        plt.imshow(img, cmap="gray_r")
    plt.show(block=True)

    tf.debugging.disable_check_numerics()


def main():
    basic_glow_Test()



from typing import Tuple

import numpy as np
import tensorflow as tf

from TFGENZOO.flows.flowbase import FlowComponent


def regular_matrix_init(shape: Tuple[int, int], dtype=None):
    """initialize with orthogonal matrix

    Sources:
        https://github.com/openai/glow/blob/master/model.py#L445-L451

    Args:
        shape: generated matrix's shape [C, C]
        dtype:

    Returns:
       np.array: w_init, orthogonal matrix [C, C]

    """
    assert len(shape) == 2, "this initialization for 2D matrix"
    assert shape[0] == shape[1], "this initialization for 2D matrix, C \times C"
    c = shape[0]
    w_init = np.linalg.qr(np.random.randn(c, c))[0].astype("float32")
    # https://github.com/jaywalnut310/glow-tts/issues/17#issuecomment-644478717
    # https://github.com/NVIDIA/waveglow/blob/d18e0f3cc2ff6bdd41244d7391140accdc41142b/glow.py#L76
    # Ensure determinant is 1.0 not -1.0
    if np.linalg.det(w_init) < 0:
        w_init[:, 0] = -1 * w_init[:, 0]
    return w_init


class Inv1x1Conv(FlowComponent):
    """Invertible 1x1 Convolution Layer

    Sources:
        https://arxiv.org/pdf/1807.03039.pdf
        https://github.com/openai/glow/blob/master/model.py#L457-L472

    Note:
   
        * forward formula
            .. math::

                \\forall i, j: z_{i, j} &= Wx_{i, j} \\\\
                LogDetJacobian &= hw \log|det(W)|\\\\
                , where &\\\\
                W &\\in \\mathbb{R}^{c \times c}\\\\
                    x &\\in \\mathbb{R}^{b \\times h\\times w \\times c}\\ \\ \\
                ({\\rm batch, height, width, channel})

        * inverse formula
            .. math::

                \\forall i, j: x_{i, j} &= W^{-1} z_{i, j}\\\\
                InverseLogDetJacobian &= - h w \log|det(W)|\\\\
                , where &\\\\
                W &\\in \\mathbb{R}^{c\\times c}\\\\
                x &\\in \\mathbb{R}^{b \\times h\\times w \\times c}\\ \\ \\
                ({\\rm batch, height, width, channel})

    Examples:

        >>> import tensorflow as tf
        >>> from TFGENZOO.flows import Inv1x1Conv
        >>> ic = Inv1x1Conv()
        >>> ic.build([None, 16, 16, 4])
        >>> ic.get_config()
        {'name': 'inv1x1_conv_1', 'trainable': {}, 'dtype': 'float32'}
        >>> inputs = tf.keras.Input([16, 16, 4])
        >>> tf.keras.Model(inputs, ic(inputs)).summary()
        Layer (type)                 Output Shape              Param #
        =================================================================
        input_3 (InputLayer)         [(None, 16, 16, 4)]       0
        _________________________________________________________________
        inv1x1_conv_1 (Inv1x1Conv)   ((None, 16, 16, 4), (None 17
        =================================================================
        Total params: 17
        Trainable params: 0
        Non-trainable params: 17
        _________________________________________________________________
    """

    def build(self, input_shape: tf.TensorShape):
        _, h, w, c = input_shape
        self.h = h
        self.w = w
        self.c = c
        self.W = self.add_weight(
            name="W",
            shape=(c, c),
            regularizer=tf.keras.regularizers.l2(0.01),
            initializer=regular_matrix_init,
        )
        super().build(input_shape)

    def __init__(self, log_det_type: str = "slogdet", **kwargs):
        super().__init__()
        self.log_det_type = log_det_type
        if self.log_det_type == "logdet":
            self.log_det_func = lambda x: tf.linalg.logdet(x)
        else:
            self.log_det_func = lambda x: tf.linalg.slogdet(x)[1]

    def get_config(self):
        config = super().get_config()
        config_update = {"log_det_type": self.log_det_type}
        config.update(config_update)
        return config

    def forward(self, x: tf.Tensor, **kwargs):
        W = self.W + tf.eye(self.c) * 1e-5
        _W = tf.reshape(W, [1, 1, self.c, self.c])
        z = tf.nn.conv2d(x, _W, [1, 1, 1, 1], "SAME")
        # scalar
        log_det_jacobian = tf.cast(
            self.log_det_func(tf.cast(W, tf.float64)) * self.h * self.w, tf.float32,
        )
        # expand as batch
        log_det_jacobian = tf.broadcast_to(log_det_jacobian, tf.shape(x)[0:1])
        return z, log_det_jacobian

    def inverse(self, z: tf.Tensor, **kwargs):
        W = self.W + tf.eye(self.c) * 1e-5
        _W = tf.reshape(tf.linalg.inv(W), [1, 1, self.c, self.c])
        x = tf.nn.conv2d(z, _W, [1, 1, 1, 1], "SAME")

        inverse_log_det_jacobian = tf.cast(
            -1 * self.log_det_func(tf.cast(W, tf.float64)) * self.h * self.w,
            tf.float32,
        )

        inverse_log_det_jacobian = tf.broadcast_to(
            inverse_log_det_jacobian, tf.shape(z)[0:1]
        )
        return x, inverse_log_det_jacobian


class Inv1x1Conv2DWithMask(FlowComponent):
    """Invertible 1x1 Convolution Layer (2D) with Mask

    Sources:
        https://arxiv.org/pdf/1807.03039.pdf
        https://github.com/openai/glow/blob/master/model.py#L457-L472

    Note:

        * forward formula
            .. math::

                \\forall i: z_{i} &= Wx_{i} \\\\
                LogDetJacobian &= t \log|det(W)|\\\\
                , where &\\\\
                W &\\in \\mathbb{R}^{c \times c}\\\\
                    x &\\in \\mathbb{R}^{b \\times h\\times w \\times c}\\ \\ \\
                ({\\rm batch, timestep, channel})

        * inverse formula
            .. math::

                \\forall i: x_{i} &= W^{-1} z_{i}\\\\
                InverseLogDetJacobian &= - t \log|det(W)|\\\\
                , where &\\\\
                W &\\in \\mathbb{R}^{c\\times c}\\\\
                x &\\in \\mathbb{R}^{b \\times t\\times c}\\ \\ \\
                ({\\rm batch, timestep, channel})

        * mask notes
            | mask shape is [B, T, M] where M may be 1
            | reference glow-tts
    """

    def __init__(self, **kwargs):
        super().__init__()

    def build(self, input_shape: tf.TensorShape):
        _, t, c = input_shape
        self.c = c
        self.W = self.add_weight(
            name="W",
            shape=(c, c),
            regularizer=tf.keras.regularizers.l2(0.01),
            initializer=regular_matrix_init,
        )
        super().build(input_shape)

    def get_config(self):
        config = super().get_config()
        config_update = {}
        config.update(config_update)
        return config

    def forward(self, x: tf.Tensor, mask: tf.Tensor = None, **kwargs):
        """
        Args:
            x    (tf.Tensor): base input tensor [B, T, C]
            mask (tf.Tensor): mask input tensor [B, T, M] where M may be 1

        Returns:
            z    (tf.Tensor): latent variable tensor [B, T, C]
            ldj  (tf.Tensor): log det jacobian [B]

        Notes:
            * mask's example
                | [[[True], [True], [True], [False],
                |  [[True], [False], [False], [False],
                |  [[True], [True], [True], [True]],
                |  [[True], [True], [True], [True]]]
        """
        # b, t, c = tf.shape(x)
        shapes = tf.shape(x)
        b = shapes[0]
        t = shapes[1]
        c = shapes[2]

        W = self.W + tf.eye(self.c) * 1e-5
        _W = tf.reshape(W, [1, self.c, self.c])
        z = tf.nn.conv1d(x, _W, [1, 1, 1], "SAME")

        # scalar
        # tf.math.log(tf.abs(tf.linalg.det(W))) == tf.linalg.slogdet(W)[1]
        log_det_jacobian = tf.cast(
            tf.linalg.slogdet(tf.cast(W, tf.float64))[1], tf.float32,
        )

        # expand as batch
        if mask is not None:
            mask_tensor = tf.cast(mask, tf.float32)
            z = z * mask_tensor
            # mask_tensor [B, T, M]
            log_det_jacobian = log_det_jacobian * tf.reduce_sum(
                tf.cast(mask, tf.float32), axis=[-2, -1]
            )
        else:
            log_det_jacobian = tf.broadcast_to(
                log_det_jacobian * tf.cast(t, tf.float32), tf.shape(x)[0:1]
            )
        return z, log_det_jacobian

    def inverse(self, z: tf.Tensor, mask: tf.Tensor = None, **kwargs):
        # b, t, c = tf.shape(x)
        shapes = tf.shape(z)
        b = shapes[0]
        t = shapes[1]
        c = shapes[2]

        W = self.W + tf.eye(self.c) * 1e-5
        _W = tf.reshape(tf.linalg.inv(W), [1, self.c, self.c])
        x = tf.nn.conv1d(z, _W, [1, 1, 1], "SAME")

        inverse_log_det_jacobian = tf.cast(
            -1 * tf.linalg.slogdet(tf.cast(W, tf.float64))[1], tf.float32,
        )

        if mask is not None:
            mask_tensor = tf.cast(mask, tf.float32)
            x = x * mask_tensor
            inverse_log_det_jacobian = inverse_log_det_jacobian * tf.reduce_sum(
                tf.cast(mask, tf.float32), axis=[-2, -1]
            )
        else:
            inverse_log_det_jacobian = tf.broadcast_to(
                inverse_log_det_jacobian * tf.cast(t, tf.float32), tf.shape(z)[0:1]
            )
        return x, inverse_log_det_jacobian



#!/usr/bin/env python3

import tensorflow as tf

from TFGENZOO.flows.flowbase import FlowComponent


class Flatten(FlowComponent):
    """Flatten Layer
    Sources:

        https://github.com/VLL-HD/FrEIA/blob/26a5d4a901831a7f0130c6059b9d50ac72ae6f47/FrEIA/modules/reshapes.py#L204-L221

    Examples:

        >>> import tenosorflow as tf
        >>> from TFGENZOO.flows import Flatten
        >>> fl = Flatten()
        >>> fl.build([None, 16, 16, 2])
        >>> fl(inputs)
        (<tf.Tensor 'flatten_2_2/Identity:0' shape=(None, 512) dtype=float32>,
         <tf.Tensor 'flatten_2_2/Identity_1:0' shape=(None,) dtype=float32>)
        >>> tf.keras.Model(inputs, fl(inputs)).summary()
        Model: "model"
        _________________________________________________________________
        Layer (type)                 Output Shape              Param #
        =================================================================
        input_1 (InputLayer)         [(None, 16, 16, 2)]       0
        _________________________________________________________________
        flatten_2 (Flatten)          ((None, 512), (None,))    1
        =================================================================
        Total params: 1
        Trainable params: 0
        Non-trainable params: 1
        _________________________________________________________________
        >>> z, ldj = fl(tf.random.normal([1024, 16, 16, 2]))
        >>> x, ildj = fl(z, invere=True)
        >>> x.shape
        TensorShape([1024, 16, 16, 2])
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.flatten = tf.keras.layers.Flatten()

    def get_config(self):
        return super().get_config()

    def build(self, input_shape):
        self.rebuild_shape = [-1] + list(input_shape)[1:]
        super().build(input_shape)

    def forward(self, x: tf.Tensor, **kwargs):
        z = self.flatten(x)
        log_det_jacobian = tf.zeros(shape=tf.shape(x)[0:1], dtype=tf.float32)
        return z, log_det_jacobian

    def inverse(self, x: tf.Tensor, **kwargs):
        z = tf.reshape(x, self.rebuild_shape)
        inverse_log_det_jacobian = tf.zeros(shape=tf.shape(x)[0:1], dtype=tf.float32)
        return z, inverse_log_det_jacobian



import tensorflow as tf

from TFGENZOO.flows.flowbase import FlowComponent


class Actnorm(FlowComponent):
    """Actnorm Layer

    Sources:

        https://github.com/openai/glow/blob/master/tfops.py#L71-L163

    Note:
   
        * initialize
            | mean = mean(first_batch)
            | var = variance(first_batch)
            | logs = log(scale / sqrt(var)) / logscale_factor
            | bias = - mean

        * forward formula
            | logs = logs * logscale_factor
            | scale = exp(logs)
            | z = (x + bias) * scale
            | log_det_jacobain = sum(logs) * H * W

        * inverse formula
            | logs = logs * logsscale_factor
            | inv_scale = exp(-logs)
            | z = x * inv_scale - bias
            | inverse_log_det_jacobian = sum(- logs) * H * W

    Attributes:

        calc_ldj: bool
            flag of calculate log det jacobian

        scale: float
            initialize batch's variance scaling
   
        logscale_factor: float
            barrier log value to - Inf

    Examples:

        >>> import tensorflow as tf
        >>> import TFGENZOO.flows import Actnorm
        >>> ac = Actnorm()
        >>> ac.build([None, 16, 16, 4])
        >>> ac.get_config()
        {'name': 'actnorm_1', ... }
        >>> inputs = tf.keras.Input([16, 16, 4])
        >>> ac(inputs)
        (<tf.Tensor 'actnorm_1_2/Identity:0' shape=(None, 16, 16, 4) dtype=float32>,
         <tf.Tensor 'actnorm_1_2/Identity_1:0' shape=(None,) dtype=float32>)
        >>>  tf.keras.Model(inputs, ac(inputs)).summary()
        Model: "model_5"
        _________________________________________________________________
        Layer (type)                 Output Shape              Param #
        =================================================================
        input_3 (InputLayer)         [(None, 16, 16, 4)]       0
        _________________________________________________________________
        actnorm_1 (Actnorm)          ((None, 16, 16, 4), (None 9
        =================================================================
        Total params: 9
        Trainable params: 0
        Non-trainable params: 9
        _________________________________________________________________

    """

    def __init__(self, scale: float = 1.0, logscale_factor: float = 3.0, **kwargs):
        super(Actnorm, self).__init__(**kwargs)
        self.scale = scale
        self.logscale_factor = logscale_factor

    def get_config(self):
        config = super().get_config()
        config_update = {"scale": self.scale, "logscale_factor": self.logscale_factor}
        config.update(config_update)
        return config

    # pylint: disable=attribute-defined-outside-init
    def build(self, input_shape: tf.TensorShape):
        if len(input_shape) == 4:
            reduce_axis = [0, 1, 2]
            b, h, w, c = list(input_shape)
            self.logdet_factor = tf.constant(h * w, dtype=tf.float32)
        else:
            raise NotImplementedError()

        self.reduce_axis = reduce_axis

        # stats_shape = [1, 1, 1, C] if input_shape == [B, H, W, C]
        stats_shape = [1 for _ in range(len(input_shape))]
        stats_shape[-1] = input_shape[-1]

        self.logs = self.add_weight(
            name="logscale",
            shape=tuple(stats_shape),
            initializer="zeros",
            trainable=True,
            aggregation=tf.VariableAggregation.MEAN,
        )
        self.bias = self.add_weight(
            name="bias",
            shape=tuple(stats_shape),
            initializer="zeros",
            trainable=True,
            aggregation=tf.VariableAggregation.MEAN,
        )

        super().build(input_shape)

    def data_dep_initialize(self, x: tf.Tensor):

        if self.initialized:
            bias, logs = self.bias, self.logs
        else:
            tf.print("initialization at {}".format(self.name))
            mean = tf.reduce_mean(x, axis=[0, 1, 2], keepdims=True)
            squared = tf.reduce_mean(tf.square(x), axis=[0, 1, 2], keepdims=True)
            variance = squared - tf.square(mean)
            bias = -mean
            logs = (
                tf.math.log(self.scale * tf.math.rsqrt(variance + 1e-6))
                / self.logscale_factor
            )

        with tf.control_dependencies([bias, logs]):
            self.bias.assign(bias)
            self.logs.assign(logs)

            super().data_dep_initialize(x)

    def forward(self, x: tf.Tensor, **kwargs):
        logs = self.logs * self.logscale_factor

        # centering
        z = x + self.bias
        # scaling
        z = z * tf.exp(logs)

        log_det_jacobian = tf.reduce_sum(logs) * self.logdet_factor
        log_det_jacobian = tf.broadcast_to(log_det_jacobian, tf.shape(x)[0:1])
        return z, log_det_jacobian

    def inverse(self, z: tf.Tensor, **kwargs):
        logs = self.logs * self.logscale_factor

        # inverse scaling
        x = z * tf.exp(-1 * logs)
        # inverse centering
        x = x - self.bias

        inverse_log_det_jacobian = -1 * tf.reduce_sum(logs) * self.logdet_factor
        inverse_log_det_jacobian = tf.broadcast_to(
            inverse_log_det_jacobian, tf.shape(x)[0:1]
        )
        return x, inverse_log_det_jacobian



#!/usr/bin/env python3

import tensorflow as tf


class GTU(tf.keras.layers.Layer):
    """GTU layer proposed in Flow-TTS

    Note:

        * formula
            .. math::

                z = tanh(W_{f, k} \star y) \odot sigmoid(W_{g, k} \star c)
    """

    def __init__(self, **kwargs):
        super().__init__()

    def build(self, input_shape: tf.TensorShape):

        self.conv_first = tf.keras.layers.Conv1D(
            input_shape[-1],
            kernel_size=1,
            strides=1,
            padding="same",
            data_format="channels_last",
        )
        self.conv_last = tf.keras.layers.Conv1D(
            input_shape[-1],
            kernel_size=1,
            strides=1,
            padding="same",
            data_format="channels_last",
        )

        super().build(input_shape)

    def get_config(self):
        config = super().get_config()
        config_update = {}
        config.update(config_update)
        return config

    def call(self, y: tf.Tensor, c: tf.Tensor, **kwargs):
        """
        Args:

            y (tf.Tensor): input contents tensor [B, T, C]
            c (tf.Tensor): input conditional tensor [B, T, C'] where C' can be different with C

        Returns:

            tf.Tensor: [B, T, C]
        """

        right = tf.nn.tanh(self.conv_first(y))
        left = tf.nn.sigmoid(self.conv_last(c))
        z = right * left
        return z


def CouplingBlock(x: tf.Tensor, cond: tf.Tensor, depth, **kwargs):
    """
    Args:

        x (tf.Tensor): input contents tensor [B, T, C]
        c (tf.Tensor): input conditional tensor [B, T, C'] where C' can be different with C
    Returns:

        tf.keras.Model: CouplingBlock
                        reference: Flow-TTS
    Examples:

        >>> import tensorflow as tf
        >>> from utils.coupling_block import CouplingBlock
        >>> x = tf.keras.layers.Input([None, 32])
        >>> c = tf.keras.layers.Input([None, 128])
        >>> cp = CouplingBlock(x, c, depth=256)
        >>> cp.summary()
        Model: "model"
        __________________________________________________________________________________________________
        Layer (type)                    Output Shape         Param #     Connected to
        ==================================================================================================
        input_1 (InputLayer)            [(None, None, 32)]   0
        __________________________________________________________________________________________________
        conv1d (Conv1D)                 (None, None, 256)    8192        input_1[0][0]
        __________________________________________________________________________________________________
        input_2 (InputLayer)            [(None, None, 128)]  0
        __________________________________________________________________________________________________
        gtu (GTU)                       (None, None, 256)    98816       conv1d[0][0]
        __________________________________________________________________________________________________
        tf_op_layer_AddV2 (TensorFlowOp [(None, None, 256)]  0           gtu[0][0]
                                                                        conv1d[0][0]
        __________________________________________________________________________________________________
        conv1d_1 (Conv1D)               (None, None, 64)     16448       tf_op_layer_AddV2[0][0]
        ==================================================================================================
        Total params: 123,456
        Trainable params: 123,456
        Non-trainable params: 0
        __________________________________________________________________________________________________
        >>> cp([x, c])
        <tf.Tensor 'model_1/Identity:0' shape=(None, None, 64) dtype=float32>
    """
    last_channel = x.shape[-1]
    c = cond

    conv1x1_1 = tf.keras.layers.Conv1D(
        depth,
        kernel_size=1,
        strides=1,
        padding="same",
        data_format="channels_last",
        use_bias=False,
        activation="relu",
    )

    gtu = GTU()

    conv1x1_2 = tf.keras.layers.Conv1D(
        last_channel * 2,
        kernel_size=1,
        strides=1,
        padding="same",
        data_format="channels_last",
        kernel_initializer="zeros",
        use_bias=False,
    )

    y = x
    y = conv1x1_1(y)
    z = gtu(y, c)
    y = tf.keras.layers.Add()([z, y])
    y = conv1x1_2(y)
    model = tf.keras.Model([x, c], y)
    return model



#!/usr/bin/env python3

from typing import Callable, Dict

import tensorflow as tf
from tensorflow.keras import layers

from TFGENZOO.flows.affine_coupling import AffineCouplingMask
from TFGENZOO.flows.flowbase import FlowComponent

Layer = layers.Layer
Conv2D = layers.Conv2D


def filter_kwargs(d: Dict):
    # utility function for Tensorflow's crasy constraint
    training = d.get("training", None)
    mask = d.get("mask", None)
    return {"training": training, "mask": mask}


class ConditionalAffineCoupling(FlowComponent):
    """Affine Coupling Layer

    Sources:
        https://github.com/masa-su/pixyz/blob/master/pixyz/flows/coupling.py

    Note:
        * forward formula
            | [x1, x2] = split(x)
            | log_scale, shift = NN([x1, c])
            | scale = sigmoid(log_scale + 2.0)
            | z1 = x1
            | z2 = (x2 + shift) * scale
            | z = concat([z1, z2])
            | LogDetJacobian = sum(log(scale))

        * inverse formula
            | [z1, z2] = split(x)
            | log_scale, shift = NN([z1, c])
            | scale = sigmoid(log_scale + 2.0)
            | x1 = z1
            | x2 = z2 / scale - shift
            | z = concat([x1, x2])
            | InverseLogDetJacobian = - sum(log(scale))

        * implementation notes
           | in Glow's Paper, scale is calculated by exp(log_scale),
           | but IN IMPLEMENTATION, scale is done by sigmoid(log_scale + 2.0)
           | where c is the conditional input for WaveGlow or cINN
           | https://arxiv.org/abs/1907.02392

        * TODO notes
           | cINN uses double coupling, but our coupling is single coupling
    """

    def __init__(
        self,
        mask_type: AffineCouplingMask = AffineCouplingMask.ChannelWise,
        scale_shift_net: Layer = None,
        scale_shift_net_template: Callable[[tf.keras.Input], tf.keras.Model] = None,
        scale_type="safe_exp",
        **kwargs
    ):
        """
        Args:
            mask_type       (AffineCouplingMask: AffineCoupling Mask type
            scale_shift_net (tf.keras.Layer): NN in the fomula (Deprecated)
            scale_shift_net_template (Callable[[tf.keras.Input], [tf.keras.Model]]): NN in the formula (for tf.keras.Model without Input Shape)
        """
        super().__init__(**kwargs)
        self.conditional_input = True
        if not scale_shift_net_template and not scale_shift_net:
            raise ValueError
        if scale_shift_net_template is not None:
            self.scale_shift_net_template = scale_shift_net_template
            self.scale_shift_net = None
        elif scale_shift_net is not None:
            self.scale_shift_net = scale_shift_net
        self.mask_type = mask_type

        self.scale_type = scale_type
        if self.scale_type not in ["safe_exp", "exp", "sigmoid"]:
            raise ValueError
        if self.scale_type == "safe_exp":
            self.scale_func = lambda log_scale: tf.exp(
                tf.clip_by_value(log_scale, -15.0, 15.0)
            )
        elif self.scale_type == "exp":
            self.scale_func = lambda log_scale: tf.exp(log_scale)
        else:
            self.scale_func = lambda log_scale: tf.nn.sigmoid(log_scale + 2.0)

    def get_config(self):
        config = super().get_config()
        config_update = {
            "scale_shit_net": self.scale_shift_net.get_config(),
            "mask_type": self.mask_type,
            "scale_type": self.scale_type,
        }
        config.update(config_update)
        return config

    def build(self, input_shape: tf.TensorShape):
        self.reduce_axis = list(range(len(input_shape)))[1:]
        if self.scale_shift_net is None:
            resnet_inputs = list(input_shape)[1:]
            resnet_inputs[-1] = int(resnet_inputs[-1] / 2)
            self.scale_shift_net = self.scale_shift_net_template(
                tf.keras.Input(resnet_inputs)
            )
        super().build(input_shape)

    def forward(self, x: tf.Tensor, cond: tf.Tensor, **kwargs):
        x1, x2 = tf.split(x, 2, axis=-1)
        z1 = x1
        h = self.scale_shift_net([x1, cond], **filter_kwargs(kwargs))
        if self.mask_type == AffineCouplingMask.ChannelWise:
            shift = h[..., 0::2]
            log_scale = h[..., 1::2]

            scale = self.scale_func(log_scale)
            z2 = (x2 + shift) * scale

            # scale's shape is [B, H, W, C]
            # log_det_jacobian's hape is [B]
            log_det_jacobian = tf.reduce_sum(tf.math.log(scale), axis=self.reduce_axis)
            return tf.concat([z1, z2], axis=-1), log_det_jacobian
        else:
            raise NotImplementedError()

    def inverse(self, z: tf.Tensor, cond: tf.Tensor, **kwargs):
        z1, z2 = tf.split(z, 2, axis=-1)
        x1 = z1
        h = self.scale_shift_net([z1, cond], **filter_kwargs(kwargs))
        if self.mask_type == AffineCouplingMask.ChannelWise:
            shift = h[..., 0::2]
            log_scale = h[..., 1::2]

            scale = self.scale_func(log_scale)
            x2 = (z2 / scale) - shift

            # scale's shape is [B, H, W, C // 2]
            # inverse_log_det_jacobian's hape is [B]
            inverse_log_det_jacobian = -1 * tf.reduce_sum(
                tf.math.log(scale), axis=self.reduce_axis
            )
            return tf.concat([x1, x2], axis=-1), inverse_log_det_jacobian
        else:
            raise NotImplementedError()


class ConditionalAffineCouplingWithMask(ConditionalAffineCoupling):
    """Conditional Affine Coupling Layer with mask

    Sources:
        https://github.com/masa-su/pixyz/blob/master/pixyz/flows/coupling.py

    Note:
        * forward formula
            | [x1, x2] = split(x)
            | log_scale, shift = NN([x1, c])
            | scale = exp(log_scale)
            | z1 = x1
            | z2 = (x2 + shift) * scale
            | z = concat([z1, z2])
            | LogDetJacobian = sum(log(scale))

        * inverse formula
            | [z1, z2] = split(x)
            | log_scale, shift = NN([z1, c])
            | scale = exp(log_scale)
            | x1 = z1
            | x2 = z2 / scale - shift
            | z = concat([x1, x2])
            | InverseLogDetJacobian = - sum(log(scale))

        * implementation notes
           | in Glow's Paper, scale is calculated by exp(log_scale),
           | but IN IMPLEMENTATION, scale is done by sigmoid(log_scale + 2.0)
           | where c is the conditional input for WaveGlow or cINN
           | https://arxiv.org/abs/1907.02392

        * TODO notes
           | cINN uses double coupling, but our coupling is single coupling
           |
           | scale > 0 because exp(x) > 0

        * mask notes
           | mask shape is [B, T, M] where M may be 1
           | reference glow-tts
    """

    def build(self, input_shape: tf.TensorShape):
        self.reduce_axis = list(range(len(input_shape)))[1:]
        if self.scale_shift_net is None:
            resnet_inputs = [None for _ in range(len(input_shape) - 1)]
            resnet_inputs[-1] = int(input_shape[-1] / 2)
            self.scale_shift_net = self.scale_shift_net_template(
                tf.keras.layers.Input(resnet_inputs)
            )
        super().build(input_shape)

    def forward(self, x: tf.Tensor, cond: tf.Tensor, mask: tf.Tensor = None, **kwargs):
        """
        Args:
            x    (tf.Tensor): base input tensor [B, T, C]
            cond (tf.Tensor): conditional input tensor [B, T, C']
            mask (tf.Tensor): mask input tensor [B, T, M] where M may be 1

        Returns:
            z    (tf.Tensor): latent variable tensor [B, T, C]
            ldj  (tf.Tensor): log det jacobian [B]

        Note:
            * mask's example
                | [[True, True, True, False],
                |  [True, False, False, False],
                |  [True, True, True, True],
                |  [True, True, True, True]]
        """
        x1, x2 = tf.split(x, 2, axis=-1)
        z1 = x1
        h = self.scale_shift_net([x1, cond], **filter_kwargs(kwargs))
        if self.mask_type == AffineCouplingMask.ChannelWise:
            shift = h[..., 0::2]
            log_scale = h[..., 1::2]

            scale = self.scale_func(log_scale)

            # apply mask into scale, shift
            # mask -> mask_tensor: [B, T] -> [B, T, 1]
            if mask is not None:
                mask_tensor = tf.cast(mask, tf.float32)
                scale *= mask_tensor
                shift *= mask_tensor
            z2 = (x2 + shift) * scale

            # scale's shape is [B, T, C]
            # log_det_jacobian's shape is [B]
            log_det_jacobian = tf.reduce_sum(tf.math.log(scale), axis=self.reduce_axis)
            return tf.concat([z1, z2], axis=-1), log_det_jacobian
        else:
            raise NotImplementedError()

    def inverse(self, z: tf.Tensor, cond: tf.Tensor, mask: tf.Tensor = None, **kwargs):
        z1, z2 = tf.split(z, 2, axis=-1)
        x1 = z1
        h = self.scale_shift_net([x1, cond], **filter_kwargs(kwargs))
        if self.mask_type == AffineCouplingMask.ChannelWise:
            shift = h[..., 0::2]
            log_scale = h[..., 1::2]

            scale = self.scale_func(log_scale)

            if mask is not None:
                mask_tensor = tf.cast(mask, tf.float32)
                scale *= mask_tensor
                shift *= mask_tensor
            x2 = (z2 / scale) - shift

            inverse_log_det_jacobian = -1 * tf.reduce_sum(
                tf.math.log(scale), axis=self.reduce_axis
            )
            return tf.concat([x1, x2], axis=-1), inverse_log_det_jacobian
        else:
            raise NotImplementedError()


class ConditionalAffineCoupling2DWithMask(ConditionalAffineCouplingWithMask):
    def build(self, input_shape: tf.TensorShape):
        super().build(input_shape)



from TFGENZOO.flows.actnorm import Actnorm
from TFGENZOO.flows.affine_coupling import AffineCoupling, AffineCouplingMask, LogScale
from TFGENZOO.flows.flatten import Flatten
from TFGENZOO.flows.flowbase import FactorOutBase, FlowComponent, FlowModule
from TFGENZOO.flows.inv1x1conv import Inv1x1Conv, regular_matrix_init
from TFGENZOO.flows.quantize import LogitifyImage
from TFGENZOO.flows.squeeze import Squeeze

__all__ = [
    "FactorOutBase",
    "FlowComponent",
    "FlowModule",
    "Actnorm",
    "AffineCouplingMask",
    "AffineCoupling",
    "LogScale",
    "Inv1x1Conv",
    "regular_matrix_init",
    "LogitifyImage",
    "Flatten",
    "Squeeze",
]



"""weight normalization layer from tensorflow addons
code is hightly borrowed from
https://github.com/tensorflow/addons/blob/v0.7.1/tensorflow_addons/layers/wrappers.py

Differences:
    initialization between multi gpu
"""
from __future__ import absolute_import, division, print_function

import logging

import tensorflow as tf


class WeightNormalization(tf.keras.layers.Wrapper):
    """This wrapper reparameterizes a layer by decoupling the weight's
    magnitude and direction.

    Note:
        This speeds up convergence by improving the
        conditioning of the optimization problem.
        Weight Normalization: A Simple Reparameterization to Accelerate
        Training of Deep Neural Networks: https://arxiv.org/abs/1602.07868
        Tim Salimans, Diederik P. Kingma (2016)
        WeightNormalization wrapper works for keras and tf layers.

    Examples:

        >>> net = WeightNormalization(
        >>>   tf.keras.layers.Conv2D(2, 2, activation='relu'),
        >>>     input_shape=(32, 32, 3),
        >>>        data_init=True)(x)
        >>> net = WeightNormalization(
        >>>   tf.keras.layers.Conv2D(16, 5, activation='relu'),
        >>>     data_init=True)(net)
        >>> net = WeightNormalization(
        >>>   tf.keras.layers.Dense(120, activation='relu'),
        >>>     data_init=True)(net)
        >>> net = WeightNormalization(
        >>>   tf.keras.layers.Dense(n_classes),
        >>>     data_init=True)(net)

    Args:
      layer (tf.keras.layers.Layer): a layer instance.
      data_init (bool): If `True` use data dependent variable initialization

    Returns:
      tf.keras.layers.Layer: Wrapped Layer
   
    Raises:
      ValueError: If not initialized with a `Layer` instance.
      ValueError: If `Layer` does not contain a `kernel` of weights
      NotImplementedError: If `data_init` is True and running graph execution
    """

    def __init__(self, layer: tf.keras.layers, data_init: bool = True, **kwargs):
        super().__init__(layer, **kwargs)
        self.data_init = data_init
        self._track_trackable(layer, name="layer")
        self.is_rnn = isinstance(self.layer, tf.keras.layers.RNN)

        if self.data_init and self.is_rnn:
            logging.warning(
                "WeightNormalization: Using `data_init=True` with RNNs "
                "is advised against by the paper. Use `data_init=False`."
            )

    def build(self, input_shape):
        """Build `Layer`"""
        input_shape = tf.TensorShape(input_shape)
        self.input_spec = tf.keras.layers.InputSpec(shape=[None] + input_shape[1:])

        if not self.layer.built:
            self.layer.build(input_shape)

        kernel_layer = self.layer.cell if self.is_rnn else self.layer

        if not hasattr(kernel_layer, "kernel"):
            raise ValueError(
                "`WeightNormalization` must wrap a layer that"
                " contains a `kernel` for weights"
            )

        if self.is_rnn:
            kernel = kernel_layer.recurrent_kernel
        else:
            kernel = kernel_layer.kernel

        # The kernel's filter or unit dimension is -1
        self.layer_depth = int(kernel.shape[-1])
        self.kernel_norm_axes = list(range(kernel.shape.rank - 1))

        self.g = self.add_weight(
            name="g",
            shape=(self.layer_depth,),
            initializer="ones",
            dtype=kernel.dtype,
            trainable=True,
        )
        self.v = kernel

        self._initialized = self.add_weight(
            name="initialized",
            shape=None,
            initializer="zeros",
            dtype=tf.dtypes.bool,
            trainable=False,
        )

        if self.data_init:
            # Used for data initialization in self._data_dep_init.
            with tf.name_scope("data_dep_init"):
                layer_config = tf.keras.layers.serialize(self.layer)
                layer_config["config"]["trainable"] = False

                self._naked_clone_layer = tf.keras.layers.deserialize(layer_config)
                self._naked_clone_layer.build(input_shape)
                self._naked_clone_layer.set_weights(self.layer.get_weights())
                if not self.is_rnn:
                    self._naked_clone_layer.activation = None

        self.built = True

    def call(self, inputs):
        """Call `Layer`"""

        def _do_nothing():
            return tf.identity(self.g)

        def _update_weights():
            # Ensure we read `self.g` after _update_weights.
            with tf.control_dependencies(self._initialize_weights(inputs)):
                return tf.identity(self.g)

        g = tf.cond(self._initialized, _do_nothing, _update_weights)

        with tf.name_scope("compute_weights"):
            # Replace kernel by normalized weight variable.
            kernel = tf.nn.l2_normalize(self.v, axis=self.kernel_norm_axes) * g

            if self.is_rnn:
                self.layer.cell.recurrent_kernel = kernel
                update_kernel = tf.identity(self.layer.cell.recurrent_kernel)
            else:
                self.layer.kernel = kernel
                update_kernel = tf.identity(self.layer.kernel)

            # Ensure we calculate result after updating kernel.
            with tf.control_dependencies([update_kernel]):
                outputs = self.layer(inputs)
                return outputs

    def compute_output_shape(self, input_shape):
        return tf.TensorShape(self.layer.compute_output_shape(input_shape).as_list())

    def _initialize_weights(self, inputs):
        """Initialize weight g.
        The initial value of g could either from the initial value in v,
        or by the input value if self.data_init is True.
        """
        with tf.control_dependencies(
            [
                tf.debugging.assert_equal(  # pylint: disable=bad-continuation
                    self._initialized, False, message="The layer has been initialized."
                )
            ]
        ):
            if self.data_init:
                assign_tensors = self._data_dep_init(inputs)
            else:
                assign_tensors = self._init_norm()
            assign_tensors.append(self._initialized.assign(True))
            return assign_tensors

    def _init_norm(self):
        """Set the weight g with the norm of the weight vector."""
        with tf.name_scope("init_norm"):
            v_flat = tf.reshape(self.v, [-1, self.layer_depth])
            v_norm = tf.linalg.norm(v_flat, axis=0)
            g_tensor = self.g.assign(tf.reshape(v_norm, (self.layer_depth,)))
            return [g_tensor]

    def _calculate_moments(self, x, axes, keep_dims=True):
        """calculate moments between multi GPU
        this function is as same as tf.nn.moments(x, axes)
        if the code is running in Single-GPU environment.
        Sources:
            https://github.com/tensorflow/tensorflow/blob/v2.2.0/tensorflow/python/keras/layers/normalization_v2.py#L165-L204
        """
        y = tf.cast(x, tf.float32) if x.dtype == tf.float16 else x
        replica_ctx = tf.distribute.get_replica_context()
        if replica_ctx:
            # define each GPU's variables
            local_sum = tf.reduce_sum(y, axis=axes, keepdims=True)
            local_squared_sum = tf.reduce_sum(tf.math.square(y), axes, keepdims=True)
            batch_size = tf.cast(tf.shape(y)[0], tf.float32)

            # reduce all GPUs' variables
            y_sum, y_squared_sum, global_batch_size = replica_ctx.all_reduce(
                tf.distribute.ReduceOp.SUM, [local_sum, local_squared_sum, batch_size]
            )

            axes_vals = [(tf.shape(y))[i] for i in range(1, len(axes))]
            multiplier = tf.cast(tf.reduce_prod(axes_vals), tf.float32)
            multiplier = multiplier * global_batch_size
            mean = y_sum / multiplier
            y_squared_mean = y_squared_sum / multiplier
            # var = E(x^2) - E(x)^2
            variance = y_squared_mean - tf.square(mean)
        else:
            mean = tf.reduce_meam(y, axes, keepdims=True)
            variance = tf.reduce_mean(
                tf.squared_difference(y, tf.stop_gradient(mean)), axes, keepdims=True
            )
        if not keep_dims:
            mean = tf.squeeze(mean, axes)
            variance = tf.squeeze(variance, axes)
        if x.dtype == tf.float16:
            return (tf.cast(mean, tf.float16), tf.cast(variance, tf.float16))
        else:
            return (mean, variance)

    def _data_dep_init(self, inputs):
        """Data dependent initialization."""
        with tf.name_scope("data_dep_init"):
            # Generate data dependent init values
            x_init = self._naked_clone_layer(inputs)
            data_norm_axes = list(range(x_init.shape.rank - 1))

            m_init, v_init = self._calculate_moments(
                x_init, data_norm_axes, keep_dims=False
            )
            scale_init = 1.0 / tf.math.sqrt(v_init + 1e-10)

            # RNNs have fused kernels that are tiled
            # Repeat scale_init to match the shape of fused kernel
            # Note: This is only to support the operation,
            # the paper advises against RNN+data_dep_init
            if scale_init.shape[0] != self.g.shape[0]:
                rep = int(self.g.shape[0] / scale_init.shape[0])
                scale_init = tf.tile(scale_init, [rep])

            # Assign data dependent init values
            g_tensor = self.g.assign(self.g * scale_init)
            if hasattr(self.layer, "bias") and self.layer.bias is not None:
                bias_tensor = self.layer.bias.assign(-m_init * scale_init)
                return [g_tensor, bias_tensor]
            else:
                return [g_tensor]

    def get_config(self):
        config = {"data_init": self.data_init}
        base_config = super().get_config()
        return {**base_config, **config}

    def remove(self):
        kernel = tf.Variable(
            tf.nn.l2_normalize(self.v, axis=self.kernel_norm_axes) * self.g,
            name="recurrent_kernel" if self.is_rnn else "kernel",
        )

        if self.is_rnn:
            self.layer.cell.recurrent_kernel = kernel
        else:
            self.layer.kernel = kernel

        return self.layer



from abc import ABCMeta, abstractmethod
from typing import List

import tensorflow as tf
from tensorflow.keras import layers

Layer = layers.Layer


class FlowBase(Layer, metaclass=ABCMeta):
    """Flow-based model's abstruct class

    Examples:

        >>> layer = FlowBase()
        >>> z = layer(x, inverse=False) # forward method
        >>> x_hat = layer(z, inverse=True) # inverse method
        >>> assert tf.reduce_sum((x - x_hat)^2) << 1e-3

    Note:

        If you need data-dependent initialization (e.g. ActNorm),
        You can write it at #initialize_parameter.

         This layer will be inheritanced by the invertible layer without log det jacobian
    """

    @abstractmethod
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.conditional_input = False

    def data_dep_initialize(self, x: tf.Tensor):
        self.initialized.assign(True)

    def build(self, input_shape: tf.TensorShape):
        self.initialized = self.add_weight(
            name="initialized",
            dtype=tf.bool,
            trainable=False,
            synchronization=tf.VariableSynchronization.ON_READ,
            aggregation=tf.VariableAggregation.ONLY_FIRST_REPLICA,
            initializer=lambda shape, dtype: False,
        )
        self.built = True

    def call(self, x: tf.Tensor, inverse=False, **kwargs):
        if inverse and not self.initialized:
            raise Exception("Invalid initialize")

        self.data_dep_initialize(x)

        if inverse:
            return self.inverse(x, **kwargs)
        else:
            return self.forward(x, **kwargs)

    @abstractmethod
    def forward(self, inputs, **kwargs):
        return inputs

    @abstractmethod
    def inverse(self, inputs, **kwargs):
        return inputs


class FlowComponent(FlowBase):
    """Flow-based model's abstruct class

    Note:
         This layer will be inheritanced by the invertible layer with log det jacobian
    """

    @abstractmethod
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    @abstractmethod
    def forward(self, x, **kwargs):
        log_det_jacobian = tf.zeros(x.shape[0:1])
        z = x
        return z, log_det_jacobian

    @abstractmethod
    def inverse(self, z, **kwargs):
        inverse_log_det_jacobian = tf.zeros(z.shape[0:1])
        x = z
        return x, inverse_log_det_jacobian

    def assert_tensor(self, x: tf.Tensor, z: tf.Tensor):
        if self.with_debug:
            tf.debugging.assert_shapes([(x, z.shape)])

    def assert_log_det_jacobian(self, log_det_jacobian: tf.Tensor):
        """assert log_det_jacobian's shape
       
        Note:
            | tf-2.0's bug
            | tf.debugging.assert_shapes([(tf.constant(1.0), (None, ))])
            | # => None (true)
            | tf.debugging.assert_shapes([(tf.constant([1.0, 1.0]), (None, ))])
            | # => None (true)
            | tf.debugging.assert_shapes([(tf.constant([[1.0], [1.0]]), (None, ))])
            | # => Error
        """
        if self.with_debug:
            tf.debugging.assert_shapes([(log_det_jacobian, (None,))])


class FlowModule(FlowBase):
    """Sequential Layer for FlowBase's Layer

    Examples:

         >>> layers = [FlowBase() for _ in range(10)]
         >>> module = FlowModule(layers)
         >>> z = module(x, inverse=False)
         >>> x_hat = module(z, inverse=True)
         >>> assert ((x - x_hat)^2) << 1e-3
    """

    def build(self, input_shape: tf.TensorShape):
        super().build(input_shape=input_shape)

    def __init__(self, components: List[FlowComponent], **kwargs):
        super().__init__(**kwargs)
        self.components = components

    def get_config(self):
        config = super().get_config()
        config_update_layer = []
        for comp in self.components:
            config_update_layer.append(comp.get_config())
        config_update = {"components", config_update_layer}
        config.update(config_update)
        return config

    def forward(self, x, **kwargs):
        z = x
        log_det_jacobian = []
        for component in self.components:
            z, ldj = component(z, inverse=False, **kwargs)
            log_det_jacobian.append(ldj)
        log_det_jacobian = sum(log_det_jacobian)
        return z, log_det_jacobian

    def inverse(self, z, **kwargs):
        x = z
        inverse_log_det_jacobian = []
        for component in reversed(self.components):
            x, ildj = component(x, inverse=True, **kwargs)
            inverse_log_det_jacobian.append(ildj)
        inverse_log_det_jacobian = sum(inverse_log_det_jacobian)
        return x, inverse_log_det_jacobian


class ConditionalFlowModule(FlowBase):
    """Sequential Layer for FlowBase's Layer

    Examples:

         >>> layers = [FlowBase() for _ in range(10)]
         >>> module = FlowModule(layers)
         >>> z = module(x, inverse=False)
         >>> x_hat = module(z, inverse=True)
         >>> assert ((x - x_hat)^2) << 1e-3
    """

    def build(self, input_shape: tf.TensorShape):
        super().build(input_shape=input_shape)
        self.conditional_input = True

    def __init__(self, components: List[FlowComponent], **kwargs):
        super().__init__(**kwargs)
        self.components = components

    def get_config(self):
        config = super().get_config()
        config_update_layer = []
        for comp in self.components:
            config_update_layer.append(comp.get_config())
        config_update = {"components", config_update_layer}
        config.update(config_update)
        return config

    def forward(self, x: tf.Tensor, cond: tf.Tensor, **kwargs):
        z = x
        log_det_jacobian = []
        for component in self.components:
            if component.conditional_input:
                z, ldj = component(z, inverse=False, cond=cond, **kwargs)
            else:
                z, ldj = component(z, inverse=False, **kwargs)
            log_det_jacobian.append(ldj)
        log_det_jacobian = sum(log_det_jacobian)
        return z, log_det_jacobian

    def inverse(self, z: tf.Tensor, cond: tf.Tensor, **kwargs):
        x = z
        inverse_log_det_jacobian = []
        for component in reversed(self.components):
            if component.conditional_input:
                x, ildj = component(x, inverse=True, cond=cond, **kwargs)
            else:
                x, ildj = component(x, inverse=True, **kwargs)
            inverse_log_det_jacobian.append(ildj)
        inverse_log_det_jacobian = sum(inverse_log_det_jacobian)
        return x, inverse_log_det_jacobian


class FactorOutBase(FlowBase):
    """Factor Out Layer in Flow-based Model

    Examples:

        >>> fo = FactorOutBase(with_zaux=False)
        >>> z, zaux = fo(x, zaux=None, inverse=False)
        >>> x = fo(z, zaux=zaux, inverse=True)

        >>> fo = FactorOutBase(with_zaux=True)
        >>> z, zaux = fo(x, zaux=zaux, inverse=False)
        >>> x, zaux = fo(z, zaux=zaux, inverse=True)
    """

    @abstractmethod
    def __init__(self, with_zaux: bool = False, **kwargs):
        super().__init__(**kwargs)
        self.with_zaux = with_zaux

    def get_config(self):
        config = super().get_config()
        config_update = {"with_zaux": self.with_zaux}
        config.update(config_update)
        return config

    def build(self, input_shape: tf.TensorShape):
        super().build(input_shape)

    def call(self, x: tf.Tensor, zaux: tf.Tensor = None, inverse=False, **kwargs):
        if inverse and not self.initialized:
            raise Exception("Invalid initialize")

        self.data_dep_initialize(x)

        if inverse:
            return self.inverse(x, zaux, **kwargs)
        else:
            return self.forward(x, zaux, **kwargs)

    @abstractmethod
    def forward(self, x: tf.Tensor, zaux: tf.Tensor, **kwargs):
        pass

    @abstractmethod
    def inverse(self, x: tf.Tensor, zaux: tf.Tensor, **kwargs):
        pass



#!/usr/bin/env python3

import tensorflow as tf

from TFGENZOO.layers.spectral_normalization import SpectralNormalization


class SpadeBN(tf.keras.layers.Layer):
    """SPADE BatchNormalization

    Sources:

        https://towardsdatascience.com/implementing-spade-using-fastai-6ad86b94030a
    """

    def __init__(self, width: int = 128, kernel_size=3, **kwargs):
        self.bn = tf.keras.layers.experimental.SyncBatchNormalization()
        self.conv0 = SpectralNormalization(
            tf.keras.layers.Conv2D(width, kernel_size=kernel_size, activation="relu")
        )
        self.conv1 = SpectralNormalization(
            tf.keras.layers.Conv2D(width, kernel_size=kernel_size, activation="relu")
        )
        self.conv2 = SpectralNormalization(
            tf.keras.layers.Conv2D(width, kernel_size=kernel_size, activation="relu")
        )

    def call(self, x: tf.Tensor, cond: tf.Tensor):
        interim_conv = self.conv0(cond)
        gamma = self.conv1(interim_conv)
        beta = self.conv2(interim_conv)
        outputs = self.bn(x) * gamma + beta
        return outputs

    def get_config(self):
        config = super().get_config()
        config_update = {"width": self.width, "kernel_size": 3}
        config.update(config_update)
        return config



from typing import Tuple

import tensorflow as tf
from tensorflow.keras import layers

from TFGENZOO.flows.utils.actnorm_activation import ActnormActivation

Layer = layers.Layer


class Conv2D(Layer):
    """Convolution layer for NHWC image

    Sources:

        https://github.com/openai/glow/blob/master/tfops.py#L235-L264

    Note:
        this layer applies

        * data-dependent normalization (actnorm, openai's Glow)

        * weight normalization for stable training

        this layer not implemented.

        * function add_edge_padding

        ref. https://github.com/openai/glow/blob/master/tfops.py#L203-L232
    """

    def __init__(
        self,
        width: int = None,
        width_scale: int = 1,
        kernel_size: Tuple[int, int] = (3, 3),
        stride: Tuple[int, int] = (1, 1),
        padding: str = "SAME",
        do_actnorm: bool = True,
        do_weightnorm: bool = False,
        initializer: tf.keras.initializers.Initializer = tf.keras.initializers.RandomNormal(
            mean=0.0, stddev=0.05
        ),
        bias_initializer: tf.keras.initializers.Initializer = "zeros",
    ):
        super().__init__()
        self.width = width
        self.width_scale = width_scale
        self.kernel_size = list(kernel_size)
        self.stride = [1] + list(stride) + [1]
        self.padding = padding
        self.do_actnorm = do_actnorm
        self.do_weightnorm = do_weightnorm
        if self.do_actnorm:
            self.activation = ActnormActivation()
        self.initializer = initializer

    def get_config(self):
        config = super().get_config()
        config_update = {
            "width": self.width,
            "width_scale": self.width_scale,
            "kernel_size": self.kernel_size,
            "stride": self.stride,
            "padding": self.padding,
            "do_actnorm": self.do_actnorm,
            "do_weightnorm": self.do_weightnorm,
            "initializer": tf.keras.initializers.serialize(
                tf.keras.initializers.get(self.initializer)
            ),
        }
        if self.do_actnorm:
            config_update.update({"activation": self.activation.get_config()})
        config.update(config_update)
        return config

    def build(self, input_shape: tf.TensorShape):
        n_in = input_shape[-1]
        n_out = self.width if self.width is not None else n_in * self.width_scale
        filters = self.kernel_size + [n_in, n_out]
        self.kernel = self.add_weight(
            name="kernel", shape=filters, dtype=tf.float32, initializer=self.initializer
        )
        self.reduce_axis = list(range(len(input_shape) - 1))
        if not self.do_actnorm:
            self.bias = self.add_weight(
                name="bias",
                shape=[1 for _ in range(len(input_shape) - 1)] + [n_out],
                initializer="zeros",
            )
        self.built = True

    def call(self, x: tf.Tensor):
        if self.do_weightnorm:
            kernel = tf.nn.l2_normalize(self.kernel, self.reduce_axis)
        else:
            kernel = self.kernel

        x = tf.nn.conv2d(x, filters=kernel, strides=self.stride, padding=self.padding)
        if self.do_actnorm:
            x = self.activation(x)
        else:
            x += self.bias
        return x



import numpy as np
import tensorflow as tf


def bits_x(
    log_likelihood: tf.Tensor, log_det_jacobian: tf.Tensor, pixels: int, n_bits: int = 8
):
    r"""bits/dims

    Sources:

        https://github.com/openai/glow/blob/master/model.py#L165-L186

    Args:
        log_likelihood (tf.Tensor): shape is [batch_size,]
        log_det_jacobian (tf.Tensor): shape is [batch_size,]
        pixels (int): e.g. HWC image => H * W * C
        n_bits (int): e.g [0 255] image => 8 = log(256)

    Returns:
        bits_x: shape is [batch_size,]

    Note:
        formula

        .. math::

          bits\_x = - \cfrac{(log\_likelihood + log\_det\_jacobian)}
          {pixels \log{2}} + n\_bits
    """
    nobj = -1.0 * (log_likelihood + log_det_jacobian)
    _bits_x = nobj / (np.log(2.0) * pixels) + n_bits
    return _bits_x


def split_feature(x: tf.Tensor, type: str = "split"):
    r"""type = [split, cross]

    TODO: implement Haar downsampling
    """
    channel = x.shape[-1]
    if type == "split":
        return x[..., : channel // 2], x[..., channel // 2 :]
    elif type == "cross":
        return x[..., 0::2], x[..., 1::2]



import numpy as np
import tensorflow as tf

from TFGENZOO.flows.flowbase import FlowBase


class LogitifyImage(FlowBase):
    """Apply Tapani Raiko's dequantization and express image in terms of logits

    Sources:

        https://github.com/taesungp/real-nvp/blob/master/real_nvp/model.py
        https://github.com/taesungp/real-nvp/blob/master/real_nvp/model.py#L42-L54
        https://github.com/tensorflow/models/blob/fe4e6b653141a197779d752b422419493e5d9128/research/real_nvp/real_nvp_multiscale_dataset.py#L1073-L1077
        https://github.com/masa-su/pixyz/blob/master/pixyz/flows/operations.py#L253-L254
        https://github.com/fmu2/realNVP/blob/8d36691df215af3678440ccb7c01a13d2b441a4a/data_utils.py#L112-L119

    Args:
        corrupution_level (float): power of added random variable.
        alpha (float)            : parameter about transform close interval to open interval
                                   [0, 1] to (1, 0)

    Note:

        We know many implementation on this quantization, but we use this formula.
        since many implementations use it.

        * forward preprocess (add noise)
            .. math::

                z &\\leftarrow 255.0 x  \\ \\because [0, 1] \\rightarrow [0, 255] \\\\
                z &\\leftarrow z + \\text{corruption_level} \\times  \\epsilon \\ where\\ \\epsilon \\sim N(0, 1)\\\\
                z &\\leftarrow z / (\\text{corruption_level} + 255.0)\\\\
                z &\\leftarrow z (1 - \\alpha) + 0.5 \\alpha \\ \\because \\ [0, 1] \\rightarrow (0, 1) \\\\
                z &\\leftarrow \log(z) - \log(1 -z)

        * forward formula
            .. math::

                 z &= logit(x (1 - \\alpha) + 0.5 \\alpha)\\\\
                   &= \\log(x) - \\log(1 - x)\\\\
                 LogDetJacobian &= sum(softplus(z) + softplus(-z) - softplus(\\log(\\cfrac{\\alpha}{1 - \\alpha})))

        * inverse formula
            .. math::

                 x &= logisitic(z)\\\\
                   &= 1 / (1 + exp( -z )) \\\\
                 x &= (x - 0.5 * \\alpha) / (1.0 -  \\alpha)\\\\
                 InverseLogDetJacobian &= sum(-2 \\log(logistic(z)) - z) + softplus(\\log(\\cfrac{\\alpha}{1 - \\alpha})))
    Examples:

        >>> import tensorflow as tf
        >>> from TFGENZOO.flows import LogitifyImage
        >>> li = LogitifyImage()
        >>> li.build([None, 32, 32, 1])
        >>> li.get_config()
        {'name': 'logitify_image_1', ...}
        >>> inputs = tf.keras.Input([32, 32, 1])
        >>> li(inputs)
        (<tf.Tensor 'logitify_image/Identity:0' shape=(None, 32, 32, 1) dtype=float32>,
        <tf.Tensor 'logitify_image/Identity_1:0' shape=(None,) dtype=float32>)
        >>> tf.keras.Model(inputs, li(inputs)).summary()
        Model: "model"
        _________________________________________________________________
        Layer (type)                 Output Shape              Param #
        =================================================================
        input_1 (InputLayer)         [(None, 32, 32, 1)]       0
        _________________________________________________________________
        logitify_image (LogitifyImag ((None, 32, 32, 1), (None 1
        =================================================================
        Total params: 1
        Trainable params: 0
        Non-trainable params: 1
        _________________________________________________________________
    """

    def build(self, input_shape: tf.TensorShape):
        super(LogitifyImage, self).build(input_shape)
        # ref. https://github.com/masa-su/pixyz/blob/master/pixyz/flows/operations.py#L254
        self.pre_logit_scale = tf.constant(
            np.log(self.alpha) - np.log(1.0 - self.alpha), dtype=tf.float32
        )
        if len(input_shape) == 4:
            self.reduce_axis = [1, 2, 3]
        elif len(input_shape) == 2:
            self.reduce_axis = [1]
        else:
            raise NotImplementedError()

        super(LogitifyImage, self).build(input_shape)

    def __init__(self, corruption_level=1.0, alpha=0.05):
        super(LogitifyImage, self).__init__()
        self.corruption_level = corruption_level
        self.alpha = alpha

    def get_config(self):
        config = super().get_config()
        config_update = {"corruption_level": self.corruption_level, "alpha": self.alpha}
        config.update(config_update)
        return config

    def forward(self, x: tf.Tensor, **kwargs):
        """
        """

        # 1. transform the domain of x from [0, 1] to [0, 255]
        z = x * 255.0

        # 2-1. add noize to pixels to dequantize them
        # and transform its domain ([0, 255]->[0, 1])
        z = z + self.corruption_level * tf.random.uniform(tf.shape(x))
        z = z / (255.0 + self.corruption_level)

        # 2-2. transform pixel values with logit to be unconstrained
        # ([0, 1]->(0, 1)).
        # TODO: Will this function polutes the image?
        z = z * (1 - self.alpha) + self.alpha * 0.5

        # 2-3. apply the logit function ((0, 1)->(-inf, inf)).
        z = tf.math.log(z) - tf.math.log(1.0 - z)

        logdet_jacobian = (
            tf.math.softplus(z)
            - tf.math.softplus(-z)
            - tf.math.softplus(self.pre_logit_scale)
        )

        logdet_jacobian = tf.reduce_sum(logdet_jacobian, self.reduce_axis)
        return z, logdet_jacobian

    def inverse(self, z: tf.Tensor, **kwargs):
        """
        """
        denominator = 1 + tf.exp(-z)
        x = 1 / denominator

        x = (x - 0.5 * self.alpha) / (1.0 - self.alpha)

        inverse_log_det_jacobian = -1 * (
            tf.math.softplus(z)
            - tf.math.softplus(-z)
            - tf.math.softplus(self.pre_logit_scale)
        )

        inverse_log_det_jacobian = tf.reduce_sum(
            inverse_log_det_jacobian, self.reduce_axis
        )

        return x, inverse_log_det_jacobian



#!/usr/bin/env python3
# reference: https://github.com/tensorflow/addons/pull/1244

import tensorflow as tf


class SpectralNormalization(tf.keras.layers.Wrapper):
    """This wrapper controls the Lipschitz constant of the layer by
    constraining its spectral norm.

    Note:
        This stabilizes the training of GANs.
        Spectral Normalization for Generative Adversarial Networks:
        https://arxiv.org/abs/1802.05957
        Takeru Miyato, Toshiki Kataoka, Masanori Koyama, Yuichi Yoshida (2018)
        SpectralNormalization wrapper works for keras and tf layers.

    Examples:

        >>> net = SpectralNormalization(
        >>>   tf.keras.layers.Conv2D(2, 2, activation="relu"),
        >>>     input_shape=(32, 32, 3))(x)
        >>> net = SpectralNormalization(
        >>>   tf.keras.layers.Conv2D(16, 5, activation="relu"))(net)
        >>> net = SpectralNormalization(
        >>>   tf.keras.layers.Dense(120, activation="relu"))(net)
        >>> net = SpectralNormalization(
        >>>   tf.keras.layers.Dense(n_classes))(net)

    Args:
      layer (tf.keras.layersLayer): a layer instance.

    Returns:
      tf.keras.layers.Layer: Wrapped Layer

    Raises:
      AssertionError: If not initialized with a `Layer` instance.
      ValueError: If initialized with negative `power_iterations`
      AttributeError: If `Layer` does not contain a `kernel` or `embeddings` of weights
    """

    def __init__(self, layer: tf.keras.layers, power_iterations: int = 1, **kwargs):
        super(SpectralNormalization, self).__init__(layer, **kwargs)
        if power_iterations <= 0:
            raise ValueError(
                "`power_iterations` should be greater than zero, got "
                "`power_iterations={}`".format(power_iterations)
            )
        self.power_iterations = power_iterations
        self._initialized = False
        self._track_trackable(layer, name="layer")

    def build(self, input_shape):
        """Build `Layer`"""

        input_shape = tf.TensorShape(input_shape).as_list()
        self.input_spec = tf.keras.layers.InputSpec(shape=input_shape)

        if not self.layer.built:
            self.layer.build(input_shape)

            if hasattr(self.layer, "kernel"):
                self.w = self.layer.kernel
            elif hasattr(self.layer, "embeddings"):
                self.w = self.layer.embeddings
            else:
                raise AttributeError(
                    "{} object has no attribute 'kernel' nor "
                    "'embeddings'".format(type(self.layer).__name__)
                )

            self.w_shape = self.w.shape.as_list()

            self.u = self.add_weight(
                shape=(1, self.w_shape[-1]),
                initializer=tf.initializers.TruncatedNormal(stddev=0.02),
                aggregation=tf.VariableAggregation.MEAN,
                trainable=False,
                name="sn_u",
                dtype=self.w.dtype,
            )

        super(SpectralNormalization, self).build()

    def call(self, inputs, training=None):
        """Call `Layer`"""
        if training is None:
            training = tf.keras.backend.learning_phase()

        if training:
            self.normalize_weights()

        output = self.layer(inputs)
        return output

    def compute_output_shape(self, input_shape):
        return tf.TensorShape(self.layer.compute_output_shape(input_shape).as_list())

    @tf.function
    def normalize_weights(self):
        """Generate spectral normalized weights.
        This method will update the value of self.w with the
        spectral normalized value, so that the layer is ready for call().
        """

        w = tf.reshape(self.w, [-1, self.w_shape[-1]])
        u = self.u

        with tf.name_scope("spectral_normalize"):
            for _ in range(self.power_iterations):
                v = tf.math.l2_normalize(tf.matmul(u, tf.transpose(w)))
                u = tf.math.l2_normalize(tf.matmul(v, w))

            sigma = tf.matmul(tf.matmul(v, w), tf.transpose(u))

            self.w.assign(self.w / sigma)
            self.u.assign(u)

    def get_config(self):
        config = {"power_iterations": self.power_iterations}
        base_config = super(SpectralNormalization, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))



import setuptools
import os

PACKAGE = "TFGENZOO"

about = {}
here = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(here, PACKAGE, "__version__.py")) as f:
    exec(f.read(), about)

setuptools.setup(
    name=PACKAGE,
    packages=setuptools.find_packages(),
    install_requires=["tensorflow>=2.1.0", "tensorflow_probability>=0.9.0", "pandas"],
    version=about["__version__"],
    author="MokkeMeguru",
    author_email="meguru.mokke@gmail.com",
    description="helper of building generative model with Tensorflow 2.x",
    long_description=(
        "heavily documented (test and description and formula)"
        + "generative model examples / layers"
    ),
    long_description_content_type="text/markdown",
    url="https://github.com/MokkeMeguru/TFGENZOO",
    licence="MIT",
    classifiers=[
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.6",
        "License :: OSI Approved :: MIT License",
        "Operating System :: POSIX :: Linux",
        "Operating System :: Microsoft :: Windows",
        "Environment :: GPU :: NVIDIA CUDA :: 10.0",
        "Environment :: GPU :: NVIDIA CUDA :: 10.1",
    ],
)



import tensorflow as tf
from tensorflow.keras import layers

Layer = layers.Layer


class ActnormActivation(Layer):
    """Actnorm Layer without inverse function

    This layer can SyncBatch Normalization, but may crash frequently.

    Sources:

        https://github.com/openai/glow/blob/master/tfops.py#L71-L163

    Attributes:
        scale (float)          : scaling
        logscale_factor (float): logscale_factor

    Note:
        * initialize
            | mean = mean(first_batch)
            | var = variance(first-batch)
            | logs = log(scale / sqrt(var)) / log-scale-factor
            | bias = -mean

        * forward formula (forward only)
            | logs = logs * log_scale_factor
            | scale = exp(logs)
            | z = (x + bias) * scale
    """

    def __init__(self, scale: float = 1.0, logscale_factor=3.0, **kwargs):
        super(ActnormActivation, self).__init__()
        self.scale = scale
        self.logscale_factor = logscale_factor

    def get_config(self):
        config = super().get_config()
        config_update = {"scale": self.scale, "logscale_factor": self.logscale_factor}
        config.update(config_update)
        return config

    def build(self, input_shape: tf.TensorShape):
        if len(input_shape) == 4:
            reduce_axis = [0, 1, 2]
            b, h, w, c = list(input_shape)
        else:
            raise NotImplementedError()

        self.reduce_axis = reduce_axis

        stats_shape = [1 for _ in range(len(input_shape))]
        stats_shape[-1] = input_shape[-1]

        # self.logs = self.add_weight(
        #     name="logscale",
        #     shape=tuple(stats_shape),
        #     initializer="zeros",
        #     trainable=True,
        # )
        # self.bias = self.add_weight(
        #     name="bias", shape=tuple(stats_shape), initializer="zeros", trainable=True
        # )

        self.mean = self.add_weight(
            name="mean",
            shape=tuple(stats_shape),
            initializer="zeros",
            trainable=True,
            aggregation=tf.VariableAggregation.MEAN,
        )
        self.squared = self.add_weight(
            name="squared",
            shape=tuple(stats_shape),
            initializer="ones",
            trainable=True,
            aggregation=tf.VariableAggregation.MEAN,
        )

        self.initialized = self.add_weight(
            name="initialized",
            dtype=tf.bool,
            trainable=False,
            initializer=lambda shape, dtype: False,
            synchronization=tf.VariableSynchronization.ON_READ,
            aggregation=tf.VariableAggregation.ONLY_FIRST_REPLICA,
        )
        self.build = True

    @property
    def bias(self):
        return -self.mean

    @property
    def logs(self):
        # var(x) = E(x^2) - E(x)^2
        variance = self.squared - tf.square(self.mean)

        logs = (
            #     # var(x) = E(x^2) - E(x)^2
            tf.math.log(self.scale * tf.math.rsqrt(variance + 1e-6))
            / self.logscale_factor
            #     variance = self.squared - tf.square(self.mean)
        )
        return logs

    def data_dep_initialize(self, x: tf.Tensor):
        if self.initialized:
            # bias, logs = self.bias, self.logs

            mean = self.mean
            squared = self.squared
        else:
            tf.print("initialization at {}".format(self.name))
            # mean, variance = tf.nn.moments(x, axes=[0, 1, 2], keepdims=True)
            # bias = -mean
            # logs = (
            #     #     # var(x) = E(x^2) - E(x)^2
            #     tf.math.log(self.scale * tf.math.rsqrt(variance + 1e-6))
            #     / self.logscale_factor
            #     #     variance = self.squared - tf.square(self.mean)
            # )

            # tf.print("initialization at {}".format(self.name))
            mean = tf.reduce_mean(x, axis=[0, 1, 2], keepdims=True)
            squared = tf.reduce_mean(tf.square(x), axis=[0, 1, 2], keepdims=True)

        with tf.control_dependencies([mean, squared]):
            self.mean.assign(mean)
            self.squared.assign(squared)
            # self.bias.assign(bias)
            # self.logs.assign(logs)
            self.initialized.assign(True)

    def call(self, x: tf.Tensor):

        self.data_dep_initialize(x)

        logs = self.logs * self.logscale_factor
        x = x + self.bias
        x = x * tf.exp(logs)
        return x


def main():
    aa = ActnormActivation()
    x = tf.keras.Input([16, 16, 2])
    y = aa(x)
    model = tf.keras.Model(x, y)
    model.summary()
    print(model.variables)
    y_ = model(tf.random.normal([32, 16, 16, 2]))
    print(y_.shape)
    print(model.variables)



import tensorflow as tf

from TFGENZOO.flows.flowbase import FlowBase


class Squeeze(FlowBase):
    """Squeeze Layer

    Sources:

        https://github.com/openai/glow/blob/master/tfops.py#L338-L352
        https://arxiv.org/pdf/1605.08803.pdf Figure 3

    Note:

        * forward formula
            | z = reshape(x, [B, H // 2, W // 2, C * 4])

        * inverse formula
            | x = reshape(z, [B, H, W, C])

        * checkerboard spacing

                e.g.

                | [[[[1], [2], [5], [6]],
                | [[3], [4], [7], [8]],
                | [[9], [10], [13], [14]],
                | [[11], [12], [15], [16]]]]

                to

                | [[[ 1,  5],
                | [ 9, 13]]]

                | [[[ 2,  6],
                | [10, 14]]]

                | [[[ 3,  7],
                | [11, 15]]]

                | [[[ 4,  8],
                | [12, 16]]]

    """

    def __init__(self, with_zaux=False):
        super().__init__()
        self.with_zaux = with_zaux

    def get_config(self):
        config = super().get_config()
        config_update = {"with_zaux": self.with_zaux}
        config.update(config_update)
        return config

    def forward(self, x: tf.Tensor, zaux: tf.Tensor = None, **kwargs):
        x = tf.nn.space_to_depth(x, 2)
        if self.with_zaux and zaux is not None:
            zaux = tf.nn.space_to_depth(zaux, 2)
            return x, zaux
        return x

    def inverse(self, z: tf.Tensor, zaux: tf.Tensor = None, **kwargs):
        z = tf.nn.depth_to_space(z, 2)
        if self.with_zaux and zaux is not None:
            zaux = tf.nn.depth_to_space(zaux, 2)
            return z, zaux
        return z


class Squeeze2DWithMask(FlowBase):
    def __init__(self, with_zaux: bool = False, n_squeeze: int = 2):
        self.with_zaux = with_zaux
        self.n_squeeze = n_squeeze
        super().__init__()

    def get_config(self):
        config = super().get_config()
        config_update = {"with_zaux": self.with_zaux, "n_squeeze": self.n_squeeze}
        config.update(config_update)
        return config

    def build(self, input_shape):
        if input_shape[-2] % self.n_squeeze != 0:
            tf.print(
                "Invalid shape size: Timestep-size {} % {} == 0".format(
                    input_shape[-2], self.n_squeeze
                )
            )
        super().build(input_shape)

    def forward(
        self, x: tf.Tensor, zaux: tf.Tensor = None, mask: tf.Tensor = None, **kwargs
    ):
        """
        Args:
            x     (tf.Tensor): input tensor [B, T, C]
            zaux  (tf.Tensor): pre-latent tensor [B, T, C'']
            mask  (tf.Tensor): mask tensor [B, T, M] where M may be 1
        Returns:
            tf.Tensor: reshaped input tensor [B, T // n_squeeze, C * 2]
            tf.Tensor: reshaped pre-latent tensor [B, T // n_squeeze, C'' * n_squeeze]
            tf.Tensor: reshaped mask tensor [B, T // 2]
        """
        # We cannot use x.shape because TF's problem.
        # b, t, c = tf.shape(x)
        shapes = tf.shape(x)
        b = shapes[0]
        t = shapes[1]
        c = shapes[2]

        t = (t // self.n_squeeze) * self.n_squeeze
        x = x[:, :t, :]  # [B, T_round, C]

        z = tf.reshape(
            tf.reshape(x, [-1, t // self.n_squeeze, self.n_squeeze, c]),
            [-1, t // self.n_squeeze, c * self.n_squeeze],
        )

        if mask is not None:
            mask = mask[:, self.n_squeeze - 1 :: self.n_squeeze, :]
            mask = tf.cast(mask, x.dtype)
        else:
            mask = tf.ones([b, t // self.n_squeeze, 1], dtype=x.dtype)

        if zaux is not None:
            _, t, c = tf.shape(zaux)
            zaux = tf.reshape(
                tf.reshape(zaux, [-1, t // self.n_squeeze, self.n_squeeze, c]),
                [-1, t // self.n_squeeze, c * self.n_squeeze],
            )
            return z * mask, mask, zaux
        else:
            return z * mask, mask

    def inverse(
        self, z: tf.Tensor, zaux: tf.Tensor = None, mask: tf.Tensor = None, **kwargs
    ):
        """
        Args:
            z    (tf.Tensor): input tensor [B, T // n_squeeze, C * n_squeeze]
            zaux (tf.Tensor): pre-latent tensor [B, T // n_squeeze, C'' * n_squeeze]
            mask (tf.Tensor): pre-latent tensor [B, T // n_squeeze, 1]
        Returns:
            tf.Tensor: reshaped input tensor [B, T, C]
            tf.Tensor: reshaped pre-latent tensor [B, T, C'']
            tf.Tensor: mask tensor [B, T, 1]
        """
        # b, t, c = tf.shape(x)
        shapes = tf.shape(x)
        b = shapes[0]
        t = shapes[1]
        c = shapes[2]

        x = tf.reshape(
            tf.reshape(z, [-1, t, self.n_squeeze, c // self.n_squeeze]),
            [-1, t * self.n_squeeze, c // self.n_squeeze],
        )

        if mask is not None:
            mask = tf.expand_dims(mask, -1)  # [B, T, 1, 1]
            mask = tf.tile(mask, [1, 1, 1, self.n_squeeze])
            mask = tf.reshape(mask, [b, t * self.n_squeeze, 1])
            mask = tf.cast(mask, dtype=x.dtype)
        else:
            mask = tf.ones([b, t * self.n_squeeze, 1], dtype=x.dtype)

        if zaux is not None:
            _, t, c = tf.shape(zaux)
            zaux = tf.reshape(
                tf.reshape(zaux, [-1, t, self.n_squeeze, c // self.n_squeeze]),
                [-1, t * self.n_squeeze, c // self.n_squeeze],
            )
            return x * mask, mask, zaux
        else:
            return x * mask, mask



import tensorflow as tf

from TFGENZOO.flows.flowbase import FactorOutBase
from TFGENZOO.flows.utils import gaussianize
from TFGENZOO.flows.utils.conv_zeros import Conv1DZeros, Conv2DZeros
from TFGENZOO.flows.utils.util import split_feature


class FactorOut(FactorOutBase):
    """Basic Factor Out Layer

    This layer drops factor-outed Tensor z_i

    Note:

        * forward procedure
           | input  : h_{i-1}
           | output : h_{i}, loss
           |
           | [z_i, h_i] = split(h_{i-1})
           |
           | loss =
           |     z_i \sim N(0, 1) if conditional is False
           |     z_i \sim N(mu, sigma) if conditional is True
           |  ,where
           | mu, sigma = Conv(h)

        * inverse procedure
           | input  : h_{i}
           | output : h_{i-1}
           |
           | sample z_i from N(0, 1) or N(mu, sigma) by conditional
           | h_{i-1} = [z_i, h_i]
    """

    def build(self, input_shape: tf.TensorShape):
        self.split_size = input_shape[-1] // 2
        super().build(input_shape)

    def __init__(
        self, with_zaux: bool = False, conditional: bool = False, dims: int = 4
    ):
        super().__init__()
        self.with_zaux = with_zaux
        self.conditional = conditional
        self.dims = dims
        if self.conditional:
            if dims == 4:
                self.conv = Conv2DZeros(width_scale=2)
            elif dims == 3:
                self.conv = Conv1DZeros(width_scale=2)

    def get_config(self):
        config = super().get_config()
        config_update = {}
        if self.conditional:
            config_update = {
                "conditional": self.conditional,
                "conv": self.conv.get_config(),
                "dims": self.dims,
            }
        else:
            config_update = {"conditional": self.conditional}
        config.update(config_update)
        return config

    def split2d_prior(self, z: tf.Tensor):
        h = self.conv(z)
        return split_feature(h, "cross")

    def calc_ll(self, z1: tf.Tensor, z2: tf.Tensor):
        if self.conditional:
            mean, logsd = self.split2d_prior(z1)
            ll = gaussianize.gaussian_likelihood(mean, logsd, z2)
        else:
            ll = gaussianize.gaussian_likelihood(
                tf.zeros(tf.shape(z2)), tf.zeros(tf.shape(z2)), z2
            )
        ll = tf.reduce_sum(ll, axis=list(range(1, len(z2.shape))))
        return ll

    def forward(self, x: tf.Tensor, zaux: tf.Tensor = None, **kwargs):
        new_z = x[..., : self.split_size]
        x = x[..., self.split_size :]

        ll = self.calc_ll(x, new_z)

        if self.with_zaux:
            zaux = tf.concat([zaux, new_z], axis=-1)
        else:
            zaux = new_z
        return x, zaux, ll

    def inverse(
        self, z: tf.Tensor, zaux: tf.Tensor = None, temparature: float = 0.2, **kwargs
    ):
        if zaux is not None:
            new_z = zaux[..., -self.split_size :]
            zaux = zaux[..., : -self.split_size]
        else:
            # TODO: sampling test
            mean, logsd = self.split2d_prior(z)
            new_z = gaussianize.gaussian_sample(mean, logsd, temparature)
        z = tf.concat([new_z, z], axis=-1)
        if self.with_zaux:
            return z, zaux
        else:
            return z


class FactorOutWithMask(FactorOut):
    """Basic Factor Out Layer With Mask

    This layer drops factor-outed Tensor z_i

    Note:

        * forward procedure
           | input  : h_{i-1}
           | output : h_{i}, loss
           |
           | [z_i, h_i] = split(h_{i-1})
           |
           | loss =
           |     z_i \sim N(0, 1) if conditional is False
           |     z_i \sim N(mu, sigma) if conditional is True
           |  ,where
           | mu, sigma = Conv(h)

        * inverse procedure
           | input  : h_{i}
           | output : h_{i-1}
           |
           | sample z_i from N(0, 1) or N(mu, sigma) by conditional
           | h_{i-1} = [z_i, h_i]

        * mask notes
            | mask shape is [B, T, M] where M may be 1
            | reference glow-tts
    """

    def calc_ll(self, z1: tf.Tensor, z2: tf.Tensor, mask_tensor: tf.Tensor = None):
        """
        Args:
           z1 (tf.Tensor): [B, T, C // 2]
           z2 (tf.Tensor): [B, T, C // 2]
        """
        with tf.name_scope("calc_log_likelihood"):
            if self.conditional:
                mean, logsd = self.split2d_prior(z1)
                ll = gaussianize.gaussian_likelihood(mean, logsd, z2)
            else:
                ll = gaussianize.gaussian_likelihood(
                    tf.zeros(tf.shape(z2)), tf.zeros(tf.shape(z2)), z2
                )
            # ll is [B, T, C // 2]
            if mask_tensor is not None:
                ll *= mask_tensor
                ll = tf.reduce_sum(ll, axis=list(range(1, len(z2.shape))))
            return ll

    def forward(self, x: tf.Tensor, zaux: tf.Tensor = None, mask=None, **kwargs):
        if mask is not None:
            mask_tensor = tf.cast(mask, tf.float32)
        else:
            mask_tensor = None
        with tf.name_scope("split"):
            new_z = x[..., : self.split_size]
            x = x[..., self.split_size :]

        ll = self.calc_ll(x, new_z, mask_tensor=mask_tensor)

        if self.with_zaux:
            zaux = tf.concat([zaux, new_z], axis=-1)
        else:
            zaux = new_z
        return x, zaux, ll

    def inverse(
        self,
        z: tf.Tensor,
        zaux: tf.Tensor = None,
        mask=None,
        temparature: float = 0.2,
        **kwargs
    ):
        if zaux is not None:
            new_z = zaux[..., -self.split_size :]
            zaux = zaux[..., : -self.split_size]
        else:
            # TODO: sampling test
            mean, logsd = self.split2d_prior(z)
            new_z = gaussianize.gaussian_sample(mean, logsd, temparature)
        z = tf.concat([new_z, z], axis=-1)
        if self.with_zaux:
            return z, zaux
        else:
            return z


class FactorOut2DWithMask(FactorOutWithMask):
    def __init__(self, with_zaux: bool = False, conditional: bool = False, **kwargs):
        super().__init__(**kwargs)
        self.with_zaux = with_zaux
        self.conditional = conditional
        if self.conditional:
            self.conv = Conv1DZeros(width_scale=2)

    def get_config(self):
        config = super().get_config()
        config_update = {}
        if self.conditional:
            config_update = {
                "conditional": self.conditional,
                "conv": self.conv.get_config(),
            }
        else:
            config_update = {"conditional": self.conditional}
        config.update(config_update)
        return config

    def split2d_prior(self, z: tf.Tensor):
        h = self.conv(z)
        return split_feature(h, "cross")

    def calc_ll(self, z1: tf.Tensor, z2: tf.Tensor, mask_tensor: tf.Tensor = None):
        """Calculate log likelihood.
        Args:
           z1 (tf.Tensor): [B, T, C // 2]
           z2 (tf.Tensor): [B, T, C // 2]
        """
        with tf.name_scope("calc_log_likelihood"):
            if self.conditional:
                mean, logsd = self.split2d_prior(z1)
                ll = gaussianize.gaussian_likelihood(mean, logsd, z2)
            else:
                ll = gaussianize.gaussian_likelihood(
                    tf.zeros(tf.shape(z2)), tf.zeros(tf.shape(z2)), z2
                )
            # ll is [B, T, C // 2]
            if mask_tensor is not None:
                ll *= mask_tensor
                ll = tf.reduce_sum(ll, axis=list(range(1, len(z2.shape))))
            return ll

    def forward(self, x: tf.Tensor, zaux: tf.Tensor = None, mask=None, **kwargs):
        if mask is not None:
            mask_tensor = tf.cast(mask, x.dtype)
        else:
            mask_tensor = None
        with tf.name_scope("split"):
            new_z = x[..., : self.split_size]
            x = x[..., self.split_size :]

        ll = self.calc_ll(x, new_z, mask_tensor=mask_tensor)

        if self.with_zaux:
            zaux = tf.concat([zaux, new_z], axis=-1)
        else:
            zaux = new_z
        return x, zaux, ll

    def inverse(
        self,
        z: tf.Tensor,
        zaux: tf.Tensor = None,
        mask=None,
        temparature: float = 0.2,
        **kwargs
    ):
        if zaux is not None:
            new_z = zaux[..., -self.split_size :]
            zaux = zaux[..., : -self.split_size]
        else:
            # TODO: sampling test
            mean, logsd = self.split2d_prior(z)
            new_z = gaussianize.gaussian_sample(mean, logsd, temparature)
        z = tf.concat([new_z, z], axis=-1)
        if self.with_zaux:
            return z, zaux
        else:
            return z


def main():
    layer = FactorOut()
    x = tf.random.normal([16, 4, 4, 128])
    y, zaux, ll = layer(x, zaux=None, inverse=False)
    _x = layer(y, zaux=zaux, inverse=True)
    print(x.shape)
    print(y.shape)
    print(zaux.shape)
    print(_x.shape)
    print(tf.reduce_sum(x - _x))
    layer = FactorOut(with_zaux=True)
    x = tf.random.normal([16, 8, 8, 8])
    zaux = tf.random.normal([16, 8, 8, 8])
    z, zaux, ll = layer(x, zaux=zaux, inverse=False)
    _x, _zaux = layer(z, zaux=zaux, inverse=True)
    print(x.shape)
    print(z.shape)
    print(zaux.shape)
    print(_x.shape)
    print(tf.reduce_mean(x - _x))
    layer = FactorOut()
    x = tf.keras.Input([8, 8, 8])
    z, zaux, ll = layer(x, zaux=None)
    model = tf.keras.Model(x, z)
    model.summary()
    return model
